"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .user import User, Broadcaster, ClientUser
from typing import TYPE_CHECKING, overload
from .errors import UnregisteredUser
from .utils import datetime_to_str
import datetime
import weakref
import asyncio

if TYPE_CHECKING:
    from .types import Data, TTMData, users, Edata, chat, channels, search, PData, streams, bits, analytics, eventsub
    from typing import List, Tuple, Literal, Callable, Any, Optional, Dict, AsyncGenerator
    from .types.eventsub import Data as EvData
    from .http import HTTPClient

import logging
_logger = logging.getLogger(__name__)

__all__ = ('ConnectionState',)


class ConnectionState:
    """
    Represents the state of the connection.
    """
    __slots__ = ('http', 'user', 'is_live', '__dispatch', '__custom_dispatch', '_events', 'ready', 'total_cost',
                 'max_total_cost', '_users', '_socket_debug', '_broadcasters', '_lock')

    def __init__(self,
                 dispatcher: Callable[..., Any],
                 custom_dispatch: Callable[..., Any],
                 http: HTTPClient,
                 socket_debug: bool) -> None:
        self.http: HTTPClient = http
        # Client User-related attributes
        self.user: Optional[ClientUser] = None
        self.is_live: Optional[bool] = None
        # Callbacks for dispatching events
        self.__dispatch: Callable[..., Any] = dispatcher
        self.__custom_dispatch: Callable[..., Any] = custom_dispatch
        # Event management
        self._events: Dict[str, Any] = {}
        self.ready: Optional[asyncio.Event] = None
        # Cost tracking
        self.total_cost: Optional[int] = None
        self.max_total_cost: Optional[int] = None
        # User management
        self._users: weakref.WeakValueDictionary[str, User] = weakref.WeakValueDictionary()
        self._broadcasters: Dict[str, Broadcaster] = {}
        # Debug and synchronization
        self._socket_debug: bool = socket_debug
        self._lock = asyncio.Lock()

    def clear(self) -> None:
        """Clears the state of the client by resetting attributes."""
        if self.ready is not None:
            self.ready.clear()
        self.user: Optional[ClientUser] = None
        self.is_live: Optional[bool] = None
        self._events: Dict[str, Any] = {}
        self._users: weakref.WeakValueDictionary[str, User] = weakref.WeakValueDictionary()
        self._broadcasters: Dict[str, Broadcaster] = {}

    def get_broadcasters(self) -> List[Broadcaster]:
        """Retrieves all broadcasters"""
        # Insure the broadcaster tokens exists.
        for user_id, broadcaster in self._broadcasters.items():
            if self.http.get_token(user_id):
                continue
            self.remove_user(user_id)

        return list(self._broadcasters.values())

    async def get_broadcaster(self, user_id: str) -> Broadcaster:
        """Retrieve a broadcaster with a given user_id, ensuring the user is authorized."""
        broadcaster = self._broadcasters.get(user_id)
        if self.http.get_token(user_id):
            if broadcaster is not None:
                return broadcaster

        await self.remove_user(user_id)
        raise UnregisteredUser('User %s is not registered. '
                               'Please register the user using `register_user`.' % user_id)

    def is_registered(self, user_id: str) -> bool:
        """checks the user is registered and authorized."""
        return self.http.get_token(user_id) is not None

    async def register_user(self,
                            access_token: Optional[str] = None,
                            refresh_token: Optional[str] = None) -> Broadcaster:
        """Registers a new user by initializing authorization and creating a broadcaster instance."""
        async with self._lock:
            data: users.OAuthToken = await self.http.initialize_authorization(access_token, refresh_token)
            self._broadcasters[data['user_id']] = Broadcaster(data['user_id'], state=self)
            self.user_register(self._broadcasters[data['user_id']])
            _logger.debug('Registered successfully. Broadcaster created for user_id: %s',
                          data['user_id'])

        return self._broadcasters[data['user_id']]

    async def remove_user(self, user_id: str) -> None:
        """Removes a registered user with its token if exists."""
        async with self._lock:
            if self._broadcasters.get(user_id):
                self._broadcasters.pop(user_id)

            if self.is_registered(user_id):
                self.http.remove_token(user_id)
                _logger.debug('Unregistered successfully. Broadcaster removed for user_id: %s', user_id)

    async def initialize_client(self, user_id: str) -> None:
        """Initializes the client with user and channel information and checks if the user is live."""
        user_data: Data[List[users.User]] = await self.http.get_users(user_id, [user_id])
        channel_data: Data[List[channels.ChannelInfo]] = await self.http.get_channel_information(user_id,
                                                                                                 [user_id])

        self.user = ClientUser(state=self, user_data=user_data['data'][0], channel_data=channel_data['data'][0])
        self._broadcasters[user_id] = self.user

        # Checks if User is Streaming.
        data: Optional[streams.StreamInfo] = await self.user.channel.stream.get_live()
        self.is_live = True if data is not None else False

    async def create_subscription(self,
                                  user_id: str,
                                  event: str,
                                  session_id: str,
                                  *,
                                  callbacks: Optional[List[Callable[..., Any]]] = None,
                                  condition_options: Optional[Dict[str, Any]] = None) -> None:
        """Creates a subscription for the given event and user, and manages event callbacks."""
        subscription: Optional[Dict[str, Any]] = self.http.get_subscription_info(event)

        if callbacks is not None and subscription is None:
            raise TypeError(f'Unknown event: `on_{event}` is not a recognized event.')

        if subscription is not None:
            async with self._lock:
                if self._events.setdefault(user_id, {}).get(subscription['name']) is None:
                    if condition_options:
                        subscription.update(subscription['name'])

                    data: TTMData[List[users.EventSubSubscription]] = await self.http.create_subscription(
                        self.user.id,
                        self.user.id,
                        user_id,
                        session_id,
                        subscription_type=subscription['name'],
                        subscription_version=subscription['version'],
                        subscription_condition=subscription['condition']
                    )
                    self._events[user_id][data['data'][0]['type']] = {
                        'id': data['data'][0]['id'],
                        'name': event,
                        'version': subscription['version'],
                        'condition_options': condition_options,
                        'callbacks': callbacks if callbacks is not None else [],
                        'auth_user_id': self.user.id
                    }
                    self.total_cost = data['total_cost']
                    self.max_total_cost = data['max_total_cost']

                    if data['total_cost'] >= 0.85 * data['max_total_cost']:
                        _logger.warning('Total cost is getting high (%s). '
                                        'Consider unsubscribing from some events.',
                                        data['total_cost'])
                else:
                    self._events[user_id][subscription['name']]['callbacks'] = list(dict.fromkeys(
                        self._events[user_id][subscription['name']]['callbacks'] + callbacks
                    ))

    async def remove_subscription(self, user_id: str, event: str) -> None:
        """Removes a subscription for the given event and user."""
        subscription: Optional[Dict[str, str]] = self.http.get_subscription_info(event)
        if subscription is not None:
            async with self._lock:
                if self._events.setdefault(user_id, {}).get(subscription['name']) is not None:
                    await self.http.delete_subscription(self._events[user_id]['auth_user_id'],
                                                        self._events[user_id][subscription['name']]['id'])
                    self._events[user_id].pop(subscription['name'])
                    if self.user.id == user_id and event in ['channel_update',
                                                             'user_update',
                                                             'stream_online',
                                                             'stream_offline']:
                        _logger.warning('Default client event `%s` removed. Unexpected behavior may occur.',
                                        event)

    def get_user(self, __id: str, /) -> User:
        try:
            return self._users[__id]
        except KeyError:
            user = User(__id, self.user.id, state=self)
            self._users[__id] = user
            return user

    async def get_users(self,
                        user_ids: Optional[List[str]] = None,
                        user_logins: Optional[List[str]] = None) -> List[User]:
        _users = []
        if user_ids is not None:
            _users = [self.get_user(user_id) for user_id in user_ids]

        if user_logins is not None and len(user_logins) >= 1:
            data: Data[List[users.User]] = await self.http.get_users(self.user.id, user_logins=user_logins)
            return [self.get_user(user['id']) for user in data['data']] + _users

        return _users

    async def get_users_chat_color(self, __users: List[User], /) -> List[chat.UserChatColor]:
        data: Data[List[chat.UserChatColor]] = await self.http.get_user_chat_color(self.user.id,
                                                                                   [u.id for u in __users])
        return data['data']

    @overload
    async def get_team_info(self,
                            team_name: Optional[str] = None,
                            team_id: Optional[str] = None) -> channels.Team:
        ...

    @overload
    async def get_team_info(self,
                            team_name: Optional[str] = None,
                            team_id: Optional[str] = None) -> None:
        ...

    async def get_team_info(self,
                            team_name: Optional[str] = None,
                            team_id: Optional[str] = None) -> Optional[channels.Team]:
        data: Data[List[channels.Team]] = await self.http.get_team_info(self.user.id, team_name, team_id)
        return data['data'][0]

    async def get_global_emotes(self) -> Tuple[List[chat.Emote], str]:
        data: Edata[List[chat.Emote]] = await self.http.get_global_emotes(self.user.id)
        return data['data'], data['template']

    async def get_emote_sets(self, emote_set_ids: List[str]) -> Tuple[List[chat.Emote], str]:
        data: Edata[List[chat.Emote]] = await self.http.get_emote_sets(self.user.id, emote_set_ids)
        return data['data'], data['template']

    async def get_global_chat_badges(self) -> List[chat.Badge]:
        data: Data[List[chat.Badge]] = await self.http.get_global_chat_badges(self.user.id)
        return data['data']

    async def get_global_cheermotes(self) -> List[bits.Cheermote]:
        data: Data[List[bits.Cheermote]] = await self.http.get_cheermotes(self.user.id)
        return data['data']

    async def fetch_channels_search(self,
                                    query: str,
                                    live_only: bool = False,
                                    first: int = 20) -> AsyncGenerator[List[search.ChannelSearch], None]:
        kwargs: Dict[str, Any] = {
            'query': query,
            'live_only': live_only,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[search.ChannelSearch]] = await self.http.search_channels(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_streams(self,
                            user_logins: Optional[List[str]] = None,
                            user_ids: Optional[List[str]] = None,
                            game_ids: Optional[List[str]] = None,
                            stream_type: Literal['all', 'live'] = 'all',
                            language: Optional[str] = None,
                            first: int = 20) -> AsyncGenerator[List[streams.StreamInfo], None]:
        kwargs: Dict[str, Any] = {
            'user_ids': user_ids,
            'user_logins': user_logins,
            'game_ids': game_ids,
            'stream_type': stream_type,
            'language': language,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[streams.StreamInfo]] = await self.http.get_streams(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_videos(self,
                           game_id: Optional[str] = None,
                           video_ids: Optional[List[str]] = None,
                           language: Optional[str] = None,
                           period: Optional[Literal['all', 'day', 'month', 'week']] = None,
                           sort: Optional[Literal['time', 'trending', 'views']] = None,
                           video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']] = None,
                           first: Optional[int] = 20) -> AsyncGenerator[List[channels.Video], None]:
        kwargs: Dict[str, Any] = {
            'video_ids': video_ids,
            'game_id': game_id,
            'language': language,
            'period': period,
            'sort': sort,
            'type': video_type,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[channels.Video]] = await self.http.get_videos(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_clips(self,
                          game_id: Optional[str] = None,
                          clip_ids: Optional[List[str]] = None,
                          started_at: Optional[str] = None,
                          ended_at: Optional[str] = None,
                          is_featured: Optional[bool] = None,
                          first: int = 20) -> AsyncGenerator[List[channels.Clip], None]:
        kwargs: Dict[str, Any] = {
            'game_id': game_id,
            'clip_ids': clip_ids,
            'started_at': datetime_to_str(started_at),
            'ended_at': datetime_to_str(ended_at),
            'first': first,
            'is_featured': is_featured,
            'after': None
        }
        while True:
            data: PData[List[channels.Clip]] = await self.http.get_clips(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def get_content_classification_labels(self, locale: streams.Locale = 'en-US') -> List[streams.CCLInfo]:
        data: Data[List[streams.CCLInfo]] = await self.http.get_content_classification_labels(self.user.id, locale)
        return data['data']

    async def fetch_top_games(self, first: int = 20) -> AsyncGenerator[List[search.Game], None]:
        kwargs: Dict[str, Any] = {
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[search.Game]] = await self.http.get_top_games(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_categories_search(self,
                                      query: str,
                                      first: int = 20) -> AsyncGenerator[List[search.CategorySearch], None]:
        kwargs: Dict[str, Any] = {
            'query': query,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[search.CategorySearch]] = await self.http.search_categories(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def get_games(self,
                        game_ids: Optional[List[str]] = None,
                        game_names: Optional[List[str]] = None,
                        igdb_ids: Optional[List[str]] = None) -> List[search.Game]:
        data: Data[List[search.Game]] = await self.http.get_games(self.user.id, game_ids, game_names, igdb_ids)
        return data['data']

    async def fetch_extension_analytics(self,
                                        extension_id: Optional[str] = None,
                                        analytics_type: Literal['overview_v2'] = 'overview_v2',
                                        started_at: Optional[datetime.datetime] = None,
                                        ended_at: Optional[datetime.datetime] = None,
                                        first: int = 20) -> AsyncGenerator[List[analytics.Extension], None]:
        kwargs: Dict[str, Any] = {
            'extension_id': extension_id,
            'analytics_type': analytics_type,
            'started_at': datetime_to_str(started_at),
            'ended_at': datetime_to_str(ended_at),
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[analytics.Extension]] = await self.http.get_extension_analytics(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_game_analytics(self,
                                   game_id: Optional[str] = None,
                                   analytics_type: Literal['overview_v2'] = 'overview_v2',
                                   started_at: Optional[datetime.datetime] = None,
                                   ended_at: Optional[datetime.datetime] = None,
                                   first: int = 20) -> AsyncGenerator[List[analytics.Game], None]:
        kwargs: Dict[str, Any] = {
            'game_id': game_id,
            'analytics_type': analytics_type,
            'started_at': datetime_to_str(started_at),
            'ended_at': datetime_to_str(ended_at),
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[analytics.Game]] = await self.http.get_game_analytics(self.user.id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def initialize_after_disconnect(self, session_id: str) -> None:
        _logger.debug('Initiating re-subscription process after disconnection for session ID: %s',
                      session_id)

        # Keeps the client updated.
        await self.initialize_client(self.user.id)

        events = self._events.copy()
        self._events = {}

        for user_id, events in events.items():
            for event_type, event_details in events.items():
                try:
                    await self.create_subscription(user_id,
                                                   event_details['name'],
                                                   session_id,
                                                   callbacks=event_details['callbacks'],
                                                   condition_options=event_details['condition_options'])
                except Exception as exc:
                    _logger.exception('Error processing event `on_%s`: %s',
                                      event_details['name'],
                                      str(exc))

    def ws_connect(self) -> None:
        self.__dispatch('connect')

    def state_ready(self) -> None:
        self.ready.set()
        self.__dispatch('ready')

    def ws_disconnect(self) -> None:
        self.ready.clear()
        self.__dispatch('disconnect')

    async def socket_raw_receive(self, data: Any):
        if self._socket_debug:
            self.__dispatch('socket_raw_receive', data)

    def user_register(self, broadcaster: Broadcaster) -> None:
        self.__dispatch('user_register', broadcaster)

    def parse(self, data: EvData[Any]) -> None:
        try:
            conditions = data['subscription']['condition']
            user_id = conditions.get('broadcaster_user_id') or conditions.get('to_broadcaster_user_id')

            event_type = data['subscription']['type']
            event = self._events.get(user_id, {}).get(event_type)

            # Incase Using CLI.
            if event is not None:
                for callback in event['callbacks']:
                    self.__custom_dispatch(event['name'], callback, data['event'])

                # Dispatch only custom events incase the user is not the client.
                if user_id != self.user.id:
                    return

            # Client events
            parse = getattr(self, 'parse_' + data['subscription']['type'].replace('.', '_'))
            parse(data['event'])
        except Exception as error:
            _logger.exception('Failed to parse event: %s', error)

    # Chat
    def parse_channel_chat_clear(self, data: eventsub.chat.ChatClearEvent) -> None:
        self.__dispatch('chat_clear', data)

    def parse_channel_chat_clear_user_messages(self, data: eventsub.chat.ClearUserMessagesEvent) -> None:
        self.__dispatch('chat_clear_user_messages', data)

    def parse_channel_chat_message(self, data: eventsub.chat.MessageEvent) -> None:
        self.__dispatch('chat_message', data)

    def parse_channel_chat_message_delete(self, data: eventsub.chat.MessageDeleteEvent) -> None:
        self.__dispatch('chat_message_delete', data)

    def parse_channel_chat_notification(self, data: eventsub.chat.NotificationEvent) -> None:
        self.__dispatch('chat_notification', data)

    def parse_channel_chat_settings_update(self, data: eventsub.chat.SettingsUpdateEvent) -> None:
        self.__dispatch('chat_settings_update', data)

    def parse_channel_shared_chat_begin(self, data: eventsub.chat.SharedChatBeginEvent) -> None:
        self.__dispatch('shared_chat_begin', data)

    def parse_channel_shared_chat_update(self, data: eventsub.chat.SharedChatUpdateEvent) -> None:
        self.__dispatch('shared_chat_update', data)

    def parse_channel_shared_chat_end(self, data: eventsub.chat.SharedChatEndEvent) -> None:
        self.__dispatch('shared_chat_end', data)

    # Bits
    def parse_channel_cheer(self, data: eventsub.bits.CheerEvent) -> None:
        self.__dispatch('cheer', data)

    # Moderation
    def parse_channel_shield_mode_begin(self, data: eventsub.moderation.ShieldModeBeginEvent) -> None:
        self.__dispatch('shield_mode_begin', data)

    def parse_channel_shield_mode_end(self, data: eventsub.moderation.ShieldModeEndEvent) -> None:
        self.__dispatch('shield_mode_end', data)

    def parse_channel_suspicious_user_message(self, data: eventsub.moderation.SuspiciousUserMessageEvent) -> None:
        self.__dispatch('suspicious_user_message', data)

    def parse_channel_suspicious_user_update(self, data: eventsub.moderation.SuspiciousUserUpdateEvent) -> None:
        self.__dispatch('suspicious_user_update', data)

    def parse_channel_warning_acknowledge(self, data: eventsub.moderation.WarningAcknowledgeEvent) -> None:
        self.__dispatch('warning_acknowledge', data)

    def parse_channel_warning_send(self, data: eventsub.moderation.WarningSendEvent) -> None:
        self.__dispatch('warning_send', data)

    def parse_automod_message_hold(self, data: eventsub.moderation.AutomodMessageHoldEvent) -> None:
        self.__dispatch('automod_message_hold', data)

    def parse_automod_message_update(self, data: eventsub.moderation.AutomodMessageUpdateEvent) -> None:
        self.__dispatch('automod_message_update', data)

    def parse_automod_settings_update(self, data: eventsub.moderation.AutomodSettingsUpdateEvent) -> None:
        self.__dispatch('automod_settings_update', data)

    def parse_automod_terms_update(self, data: eventsub.moderation.AutomodTermsUpdateEvent) -> None:
        self.__dispatch('automod_terms_update', data)

    def parse_channel_ban(self, data: eventsub.moderation.BanEvent) -> None:
        self.__dispatch('ban', data)

    def parse_channel_unban(self, data: eventsub.moderation.UnbanEvent) -> None:
        self.__dispatch('unban', data)

    def parse_channel_unban_request_create(self, data: eventsub.moderation.UnbanRequestCreateEvent) -> None:
        self.__dispatch('unban_request_create', data)

    def parse_channel_moderate(self, data: eventsub.moderation.UnbanRequestResolveEvent) -> None:
        self.__dispatch('channel_moderate', data)

    def parse_channel_unban_request_resolve(self, data: eventsub.moderation.ModerateEvent) -> None:
        self.__dispatch('unban_request_resolve', data)

    def parse_channel_moderator_add(self, data: eventsub.moderation.ModeratorAddEvent) -> None:
        self.__dispatch('moderator_add', data)

    def parse_channel_moderator_remove(self, data: eventsub.moderation.ModeratorRemoveEvent) -> None:
        self.__dispatch('moderator_remove', data)

    # Channels
    def parse_channel_update(self, data: eventsub.channels.ChannelUpdateEvent) -> None:
        self.user.channel.title = data['title']
        self.user.channel.language = data['language']
        self.user.channel.category_id = data['category_id']
        self.user.channel.category_name = data['category_name']
        self.user.channel.ccl = data['content_classification_labels']
        self.__dispatch('channel_update', data)

    def parse_channel_follow(self, data: eventsub.channels.FollowEvent) -> None:
        self.__dispatch('follow', data)

    def parse_channel_subscribe(self, data: eventsub.channels.SubscribeEvent) -> None:
        self.__dispatch('subscribe', data)

    def parse_channel_subscription_end(self, data: eventsub.channels.SubscriptionEndEvent) -> None:
        self.__dispatch('subscription_end', data)

    def parse_channel_subscription_gift(self, data: eventsub.channels.SubscriptionGiftEvent) -> None:
        self.__dispatch('subscription_gift', data)

    def parse_channel_subscription_message(self, data: eventsub.channels.SubscriptionMessageEvent) -> None:
        self.__dispatch('subscription_message', data)

    def parse_channel_vip_add(self, data: eventsub.channels.VIPAddEvent) -> None:
        self.__dispatch('vip_add', data)

    def parse_channel_vip_remove(self, data: eventsub.channels.VIPRemoveEvent) -> None:
        self.__dispatch('vip_remove', data)

    # Interaction
    def parse_channel_channel_points_automatic_reward_redemption_add(
            self, data: eventsub.interaction.AutomaticRewardRedemptionAddEvent) -> None:
        self.__dispatch('points_automatic_reward_redemption_add', data)

    def parse_channel_channel_points_custom_reward_add(self, data: eventsub.interaction.PointRewardEvent) -> None:
        self.__dispatch('points_reward_add', data)

    def parse_channel_channel_points_custom_reward_update(self,
                                                          data: eventsub.interaction.PointRewardEvent) -> None:
        self.__dispatch('points_reward_update', data)

    def parse_channel_channel_points_custom_reward_remove(self,
                                                          data: eventsub.interaction.PointRewardEvent) -> None:
        self.__dispatch('points_reward_remove', data)

    def parse_channel_channel_points_custom_reward_redemption_add(
            self, data: eventsub.interaction.RewardRedemptionEvent) -> None:
        self.__dispatch('points_reward_redemption_add', data)

    def parse_channel_channel_points_custom_reward_redemption_update(
            self, data: eventsub.interaction.RewardRedemptionEvent) -> None:
        self.__dispatch('points_reward_redemption_update', data)

    def parse_channel_poll_begin(self, data: eventsub.interaction.PollBeginEvent) -> None:
        self.__dispatch('poll_begin', data)

    def parse_channel_poll_progress(self, data: eventsub.interaction.PollProgressEvent) -> None:
        self.__dispatch('poll_progress', data)

    def parse_channel_poll_end(self, data: eventsub.interaction.PollEndEvent) -> None:
        self.__dispatch('poll_end', data)

    def parse_channel_prediction_begin(self, data: eventsub.interaction.PredictionBeginEvent) -> None:
        self.__dispatch('prediction_begin', data)

    def parse_channel_prediction_progress(self, data: eventsub.interaction.PredictionProgressEvent) -> None:
        self.__dispatch('prediction_progress', data)

    def parse_channel_prediction_lock(self, data: eventsub.interaction.PredictionLockEvent) -> None:
        self.__dispatch('prediction_lock', data)

    def parse_channel_prediction_end(self, data: eventsub.interaction.PredictionEndEvent) -> None:
        self.__dispatch('prediction_end', data)

    def parse_channel_hype_train_begin(self, data: eventsub.interaction.HypeTrainEvent) -> None:
        self.__dispatch('hype_train_begin', data)

    def parse_channel_hype_train_progress(self, data: eventsub.interaction.HypeTrainEvent) -> None:
        self.__dispatch('hype_train_progress', data)

    def parse_channel_hype_train_end(self, data: eventsub.interaction.HypeTrainEndEvent) -> None:
        self.__dispatch('hype_train_end', data)

    # Activity
    def parse_channel_charity_campaign_donate(self, data: eventsub.activity.CharityDonationEvent) -> None:
        self.__dispatch('charity_campaign_donate', data)

    def parse_channel_charity_campaign_start(self, data: eventsub.activity.CharityCampaignStartEvent) -> None:
        self.__dispatch('charity_campaign_start', data)

    def parse_channel_charity_campaign_progress(self, data: eventsub.activity.CharityCampaignProgressEvent) -> None:
        self.__dispatch('charity_campaign_progress', data)

    def parse_channel_charity_campaign_stop(self, data: eventsub.activity.CharityCampaignStopEvent) -> None:
        self.__dispatch('charity_campaign_stop', data)

    def parse_channel_goal_begin(self, data: eventsub.activity.GoalsEvent) -> None:
        self.__dispatch('goal_begin', data)

    def parse_channel_goal_progress(self, data: eventsub.activity.GoalsEvent) -> None:
        self.__dispatch('goal_progress', data)

    def parse_channel_goal_end(self, data: eventsub.activity.GoalsEvent) -> None:
        self.__dispatch('goal_end', data)

    # Streams
    def parse_channel_ad_break_begin(self, data: eventsub.streams.AdBreakBeginEvent) -> None:
        self.__dispatch('ad_break_begin', data)

    def parse_channel_raid(self, data: eventsub.streams.RaidEvent) -> None:
        self.__dispatch('raid', data)

    def parse_channel_shoutout_create(self, data: eventsub.streams.ShoutoutCreateEvent) -> None:
        self.__dispatch('shoutout_create', data)

    def parse_channel_shoutout_receive(self, data: eventsub.streams.ShoutoutReceivedEvent) -> None:
        self.__dispatch('shoutout_received', data)

    def parse_stream_online(self, data: eventsub.streams.StreamOnlineEvent) -> None:
        self.is_live = True
        self.__dispatch('stream_online', data)

    def parse_stream_offline(self, data: eventsub.streams.StreamOfflineEvent) -> None:
        self.is_live = False
        self.__dispatch('stream_offline', data)

    # Users
    def parse_user_update(self, data: eventsub.users.UserUpdateEvent) -> None:
        self.user.name = data['user_login']
        self.user.display_name = data['user_name']
        self.user.description = data['description']
        self.user.email = data.get('email') or None
        self.__dispatch('user_update', data)

    def parse_user_whisper_message(self, data: eventsub.users.WhisperReceivedEvent) -> None:
        self.__dispatch('whisper_received', data)

    def parse_channel_chat_user_message_hold(self, data: eventsub.users.MessageHoldEvent) -> None:
        self.__dispatch('chat_user_message_hold', data)

    def parse_channel_chat_user_message_update(self, data: eventsub.users.MessageUpdateEvent) -> None:
        self.__dispatch('chat_user_message_update', data)



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .gateway import EventSubWebSocket, ReconnectWebSocket
from .utils import setup_logging, ExponentialBackoff
from .errors import HTTPException, ConnectionClosed
from .state import ConnectionState
from typing import TYPE_CHECKING
from .http import HTTPClient
import datetime
import asyncio
import aiohttp

if TYPE_CHECKING:
    from typing import Optional, Type, Self, Callable, Any, List, Tuple, Dict, AsyncGenerator, Literal
    from .types import chat, channels, search, streams, bits, analytics, users
    from .user import User, ClientUser
    from .channel import ClientChannel
    from types import TracebackType

import logging
_logger = logging.getLogger(__name__)

__all__ = ('Client',)


class Client:
    """
    A client for interacting with the Twitch API.

    The Client class enables interaction with the Twitch API. After authentication,
    it manages HTTP requests and WebSocket connections for real-time updates,
    providing access to a broad range of Helix API endpoints.

    Parameters
    ----------
    client_id: str
        The client ID used for authentication with the Twitch API.
    client_secret: Optional[str]
        The client secret used for authentication with the Twitch API. Default is None.
    **options:
        Additional configuration options:

        - cli: [bool][bool]

            Flag indicating if CLI mode is enabled.

            **Default**: `False`

            ___

        - cli_port: [int][]

            The port number used for CLI mode.

            **Default**: `8080`
            ___

        - socket_debug: [False][bool]

            Flag indicating if raw WebSocket messages should be dispatched for debugging purposes.
            If enabled, raw WebSocket messages are dispatched to the on_socket_raw_receive.

            **Default**: `False`
            ___

        - proxy: [Optional][typing.Optional][[str][str]]

            The proxy URL to use for HTTP requests.

            **Default**: `None`
            ___

        - proxy_auth: [Optional][typing.Optional][aiohttp.BasicAuth]

            Authentication details for the proxy, if required.

            **Default**: `None`
    """

    def __init__(self, client_id: str, client_secret: Optional[str] = None, **options) -> None:
        cli: bool = options.get('cli', False)
        cli_port: int = options.get('cli_port', 8080)
        _socket_debug: bool = options.get('socket_debug', False)
        proxy: Optional[str] = options.pop('proxy', None)
        proxy_auth: Optional[aiohttp.BasicAuth] = options.pop('proxy_auth', None)

        self.client_id: str = client_id
        self.client_secret: str = client_secret

        self.loop: Optional[asyncio.AbstractEventLoop] = None
        self.http: HTTPClient = HTTPClient(client_id, client_secret,
                                           loop=self.loop,
                                           proxy=proxy,
                                           proxy_auth=proxy_auth,
                                           cli=cli,
                                           cli_port=cli_port)

        self._connection: ConnectionState = ConnectionState(
            dispatcher=self.dispatch,
            custom_dispatch=self.custom_dispatch,
            http=self.http,
            socket_debug=_socket_debug
        )
        self._closing_task: Optional[asyncio.Task] = None
        self.ws: Optional[EventSubWebSocket] = None

    @property
    def user(self) -> Optional[ClientUser]:
        """
        Retrieve the current user associated with the client.

        Returns
        -------
        Optional[ClientUser]
            The current user if authenticated, otherwise None.
        """
        return self._connection.user

    @property
    def channel(self) -> Optional[ClientChannel]:
        """
        Retrieve the channel of the current user associated with the client.

        Returns
        -------
        Optional[ClientChannel]
            The channel of the current user if authenticated, otherwise None.
        """
        return self._connection.user.channel if self._connection.user is not None else None

    @property
    def total_subscription_cost(self) -> Optional[int]:
        """
        Retrieve the total cost of all active subscriptions for the current user.

        Returns
        -------
        Optional[int]
            The total cost of all active subscriptions if the user is authenticated, otherwise None.
        """
        return self._connection.total_cost

    @property
    def max_subscription_cost(self) -> Optional[int]:
        """
        Retrieve the maximum allowed total cost for all subscriptions.

        Returns
        -------
        Optional[int]
            The maximum allowed total cost for all subscriptions if the user is authenticated, otherwise None.
        """
        return self._connection.max_total_cost

    async def close(self) -> None:
        """
        Closes the connection to the Twitch API.

        This method ensures that all connections, including WebSocket and HTTP connections,
        are properly closed. If a closing task is already running, it waits for it to complete.

        If WebSocket is open, it is closed with a normal closure code (1000). After closing
        the WebSocket, it clears the connection state and closes the HTTP client.
        """
        if self._closing_task:
            return await self._closing_task

        async def _close():
            if self.ws is not None and self.ws.is_open:
                await self.ws.close(code=1000)

            self.clear()
            await self.http.close()
            self.loop = None

        self._closing_task = asyncio.create_task(_close())
        await self._closing_task

    def clear(self) -> None:
        """
        Clears the connection state and resets the closing task.

        This method is used to reset the connection state and prepare the client for a fresh
        start or clean up after closing the connection.
        """
        self._connection.clear()
        self._closing_task = None

    def is_closed(self) -> bool:
        """
        Check if the client is currently closing or closed.

        Returns
        -------
        bool
            True if the client is in the process of closing or has been closed, otherwise False.
        """
        return self._closing_task is not None

    async def __aenter__(self) -> Self:
        """
        Asynchronous context manager entry method.

        Returns
        -------
        Self
            The current instance of the client, allowing it to be used within an async context manager.
        """
        return self

    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_value: Optional[BaseException],
                        traceback: Optional[TracebackType]) -> None:
        """
        Asynchronous context manager exit method.

        Closes the HTTP client connection when exiting the async context manager.

        Parameters
        ----------
        exc_type: Optional[Type[BaseException]]
            The type of the exception raised, if any.
        exc_value: Optional[BaseException]
            The exception instance, if any.
        traceback: Optional[TracebackType]
            The traceback object, if any.
        """
        if self._closing_task:
            await self._closing_task
        else:
            await self.close()

    async def wait_until_ready(self) -> None:
        """
        Wait until the client is ready.
        """
        if self._connection.ready is None:
            raise RuntimeError(
                'Client not initialized. Ensure authorization or context manager is used before this call.'
            )
        await self._connection.ready.wait()

    async def _async_loop(self) -> None:
        # Starts the asynchronous loop for managing client operations.
        loop = asyncio.get_running_loop()
        self.loop = loop
        self.http.loop = loop

    async def setup_hook(self) -> None:
        """
        Perform additional setup before the client is ready.

        ???+ Warning
            Do not use `wait_until_ready()` within this method as it may cause
            it to freeze.

        You can configure or set up additional extensions or services as required.
        """
        pass

    @staticmethod
    async def on_error(event_name: str, error: Exception, /, *args: Any, **kwargs: Any) -> None:
        """
        Handle errors occurring during event dispatch.

        This static method logs an exception that occurred during the processing of an event.

        Parameters
        ----------
        event_name: str
            The name of the event that caused the error.
        error: Exception
            The exception that was raised.
        *args: Any
            Positional arguments passed to the event.
        **kwargs: Any
            Keyword arguments passed to the event.
        """
        _logger.exception('Ignoring error: %s from %s, args: %s kwargs: %s', error, event_name,
                          args, kwargs)

    async def _run_event(self, coro: Callable[..., Any], event_name: str, *args: Any, **kwargs: Any) -> None:
        # Run an event coroutine and handle exceptions.
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as error:
            await self.on_error(event_name, error, *args, **kwargs)

    def custom_dispatch(self, event: str, coro: Callable[..., Any], /, *args: Any, **kwargs: Any) -> None:
        # Dispatch a custom event with a coroutine callback.
        try:
            if asyncio.iscoroutinefunction(coro):
                _logger.debug('Dispatching custom event %s', event)
                wrapped = self._run_event(coro, event, *args, **kwargs)
                # Schedule the task
                self.loop.create_task(wrapped, name=f'twitch:custom:{event}')
        except AttributeError:
            pass
        except Exception as error:
            _logger.error('Event: %s Error: %s', event, error)

    def dispatch(self, event: str, /, *args: Any, **kwargs: Any) -> None:
        # Dispatch a specified event with a coroutine callback.
        method = 'on_' + event
        try:
            coro = getattr(self, method)
            if coro is not None and asyncio.iscoroutinefunction(coro):
                _logger.debug('Dispatching event %s', event)
                wrapped = self._run_event(coro, method, *args, **kwargs)
                # Schedule the task
                self.loop.create_task(wrapped, name=f'twitch:{method}')
        except AttributeError:
            pass
        except Exception as error:
            _logger.error('Event: %s Error: %s', event, error)

    def event(self, coro: Callable[..., Any], /) -> None:
        """
        Register a coroutine function as an event handler.

        This method assigns the given coroutine function to be used as an event handler with the same
        name as the coroutine function.

        Parameters
        ----------
        coro: Callable[..., Any]
            The coroutine function to register as an event handler.

        Example
        -------
        ```py
        @client.event
        async def on_ready():
            print('Ready!')
        ```
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The registered event must be a coroutine function')
        setattr(self, coro.__name__, coro)

    async def add_custom_event(self,
                               name: str,
                               /,
                               user: User,
                               callback: Callable[..., Any],
                               *,
                               options: Optional[Dict[str, Any]] = None
                               ) -> None:
        """
        Add a custom event for a user.

        This method registers a callback function for a custom event and user.

        ???+ Warning
             You must be at least a moderator on their channel to subscribe to events on behalf of other users.

        Parameters
        ----------
        name: str
            The name of the event to subscribe to.
        user: User
            The user for whom the subscription is being created.
        callback: Callable[..., Any]
            The coroutine function to call when the event occurs.
        options: Optional[Dict[str, Any]]
            Custom event conditions.

        Example
        --------
        ```py
        async def on_yonk(data: eventsub.chat.MessageEvent):
            print(data)

        async def on_random_channel_message(data: eventsub.chat.MessageEvent):
            print(data)

        @client.event
        async def on_ready():
            user = await client.get_user('snifo')
            await client.add_custom_event('on_chat_message', user, on_random_channel_message)

            # Options not required, but sometimes we want to listen to a specific reward.
            await client.add_custom_event('on_points_reward_redemption_add', user, on_yonk, options={'reward_id': '1'})
        ```
        """
        if not asyncio.iscoroutinefunction(callback):
            raise TypeError('The registered custom event must be a coroutine function')

        if not name.startswith('on_'):
            raise TypeError('Event name must start with "on_" to be recognized as a valid custom event.')

        await self._connection.create_subscription(user.id,
                                                   name.replace('on_', '', 1),
                                                   self.ws.session_id,
                                                   callbacks=[callback],
                                                   condition_options=options)

    async def remove_custom_event(self, name: str, /, user: User) -> None:
        """
        Remove a custom event for a user.

        This method unsubscribes a user from a custom event.

        Parameters
        ----------
        name: str
            The name of the event to unsubscribe from.
        user: User
            The user whose subscription is being removed.
        """

        if not name.startswith('on_'):
            raise TypeError('Event name must start with "on_" to be recognized as a valid custom event.')

        await self._connection.remove_subscription(user.id, name.replace('on_', '', 1))

    async def authorize(self, access_token: str, refresh_token: Optional[str] = None) -> None:
        """
        Authorize the client with the given access and optionally refresh token.

        This method initializes the event loop if it's not already running,
        then calls the HTTP clients authorize method to authenticate with
        the Twitch API. After successful authorization, it initializes the
        client connection state with the user ID.

        Parameters
        ----------
        access_token: str
            The OAuth2 access token used for authentication.
        refresh_token: Optional[str], default=None
            The OAuth2 refresh token used to obtain a new access token if needed.
        """
        if self.loop is None:
            await self._async_loop()
        self._connection.ready = asyncio.Event()
        data: users.OAuthToken = await self.http.initialize_authorization(access_token, refresh_token)
        await self._connection.initialize_client(user_id=data['user_id'])
        await self.setup_hook()

    async def connect(self, *, reconnect: bool = True) -> None:
        """
        Establish a WebSocket connection to the Twitch API and handle events.

        This method tries to connect to the Twitch WebSocket server and handle
        incoming events. If the WebSocket connection is closed or errors occur,
        it handles reconnections based on the `reconnect` flag. It also logs
        connection errors and attempts to reconnect if necessary.

        Parameters
        ----------
        reconnect: bool
            Indicates whether to attempt reconnection if the WebSocket connection is lost.
        """
        kwargs: Dict[str, Any] = {'resume': False, 'initial': True}

        backoff = ExponentialBackoff()

        while not self.is_closed():
            try:
                websocket = EventSubWebSocket.initialize_websocket(self, self._connection, **kwargs)
                self.ws = await asyncio.wait_for(websocket, timeout=60.0)
                kwargs['initial'] = False
                while True:
                    await self.ws.poll_handle_dispatch()
            except ReconnectWebSocket as exc:
                _logger.debug('Websocket is reconnecting to %s', exc.url)
                kwargs['gateway'] = exc.url
                kwargs['resume'] = True

            except (OSError, HTTPException, ConnectionClosed, asyncio.TimeoutError, aiohttp.ClientError) as exc:
                self._connection.ws_disconnect()
                if not reconnect:
                    await self.close()
                    if isinstance(exc, ConnectionClosed) and exc.code == 1000:
                        return
                    raise

                if self.is_closed():
                    return

                delay = backoff.get_delay()
                _logger.exception('Attempting a reconnect in %d seconds.', delay)
                await asyncio.sleep(delay)
                kwargs['initial'] = False
                continue

    async def start(self, access_token: str, refresh_token: Optional[str] = None, *, reconnect: bool = True) -> None:
        """
        Start the client by authorizing and connecting to the Twitch API.

        This method authorizes the client with the provided access token and
        optionally refresh token. It then establishes a WebSocket connection
        to the Twitch API and starts handling events. It raises errors if the
        access token or refresh token is missing based on the client configuration.

        Parameters
        ----------
        access_token: str
            The OAuth2 access token used for authentication.
        refresh_token: Optional[str]
            The OAuth2 refresh token used to obtain a new access token if needed.
        reconnect: bool
            Indicates whether to attempt reconnection if the WebSocket connection is lost.

        Raises
        ------
        TypeError
            If the access token or refresh token is missing or if the client secret is not provided when needed.
        """
        if isinstance(self, Client) and access_token is None:
            if refresh_token and not self.client_secret:
                raise TypeError('Missing client secret.')

            if not refresh_token:
                raise TypeError('Missing client access token or refresh token.')
        await self.authorize(access_token, refresh_token)
        await self.connect(reconnect=reconnect)

    def run(self,
            access_token: Optional[str] = None,
            refresh_token: Optional[str] = None,
            *,
            reconnect: bool = True,
            log_handler: Optional[logging.Handler] = None,
            log_level: Optional[int] = None,
            root_logger: bool = False) -> None:
        """
        Start the client and run it until interrupted.

        !!! danger
            This function must be the last function to call as it blocks
            the execution of anything after it.

        This method sets up logging, initializes the client, and starts the main
        asynchronous process. The client will run and handle events until the
        process is interrupted (e.g., by a KeyboardInterrupt). The logging setup
        is customizable via parameters.

        Parameters
        ----------
        access_token: Optional[str]
            The OAuth2 access token used for authentication. If None, the client
            will need to obtain it from another source.
        refresh_token: Optional[str]
            The OAuth2 refresh token used to obtain a new access token if needed.
            If None, the client will need to obtain it from another source.
        reconnect: bool
            Indicates whether to attempt reconnection if the WebSocket connection is lost.
        log_handler: Optional[logging.Handler]
            A logging handler to be used for logging output. If None, a default handler will be set up.
        log_level: Optional[int]
            The logging level to be used. If None, a default level will be used.
        root_logger: bool
            If True, the logging configuration will apply to the root logger. Otherwise, it applies to a new logger.
        """
        if log_handler is None:
            setup_logging(handler=log_handler, level=log_level, root=root_logger)

        async def runner() -> None:
            """
            Inner function to run the main process asynchronously.
            """
            async with self:
                await self.start(access_token, refresh_token, reconnect=reconnect)

        try:
            asyncio.run(runner())
        except KeyboardInterrupt:
            return

    async def get_user(self, name: str) -> Optional[User]:
        """
        Retrieve a user by their login name.

        Parameters
        ----------
        name: str
            The login name of the user to retrieve.

        Returns
        -------
        Optional[User]
            The User object if found; otherwise, None.
        """
        data: List[User] = await self._connection.get_users(user_logins=[name])
        return data[0] if len(data) != 0 else None

    def get_user_by_id(self, __id: str, /) -> User:
        """
        Retrieve a user by their ID.

        ???+ note
            This method directly creates a User object with the provided ID, making it faster
            as it avoids an additional API request to retrieve the user ID.

        Parameters
        ----------
        __id: str
            The ID of the user to retrieve.

        Returns
        -------
        User
            The User object initialized with the given ID.
        """
        return self._connection.get_user(__id)

    async def get_users(self,
                        names: Optional[List[str]] = None,
                        ids: Optional[List[str]] = None) -> List[User]:
        """
        Retrieve multiple users by their names or IDs.

        ???+ note
            Using IDs is efficient as it avoids extra requests. If names are provided,
            it makes a single request to fetch all users by name.

        Parameters
        ----------
        names: Optional[List[str]]
            A list of user login names to retrieve.
        ids: Optional[List[str]]
            A list of user IDs to retrieve.

        Returns
        -------
        List[User]
            A list of User objects corresponding to the provided names or IDs.
        """
        data: List[User] = await self._connection.get_users(ids, names)
        return data

    async def get_users_chat_color(self, __users: List[User], /) -> List[chat.UserChatColor]:
        """
        Retrieve chat color information for a list of users.

        Parameters
        ----------
        __users: List[User]
            A list of User objects for which to retrieve chat color settings.

        Returns
        -------
        List[chat.UserChatColor]
            A list of dictionaries where each dictionary contains users chat color settings.
        """
        data: List[chat.UserChatColor] = await self._connection.get_users_chat_color(__users)
        return data

    async def get_team(self, name: str) -> channels.Team:
        """
        Retrieve a team by its name.

        Parameters
        ----------
        name: str
            The name of the team to retrieve.

        Returns
        -------
        channels.Team
            A dictionary representing the team's information.
        """
        data: Optional[channels.Team] = await self._connection.get_team_info(team_name=name)
        return data

    async def get_team_by_id(self, __id: str, /) -> channels.Team:
        """
        Retrieve a team by its ID.

        Parameters
        ----------
        __id: str
            The ID of the team to retrieve.

        Returns
        -------
        Optional[channels.Team]
            A dictionary representing the team's information.
        """
        data: Optional[channels.Team] = await self._connection.get_team_info(team_id=__id)
        return data

    async def get_global_emotes(self) -> Tuple[List[chat.Emote], str]:
        """
        Retrieve global emotes and the template URL.

        Returns
        -------
        Tuple[List[chat.Emote], str]
            A tuple containing a list of dictionaries representing global emotes and a template URL string.
        """
        data: Tuple[List[chat.Emote], str] = await self._connection.get_global_emotes()
        return data

    async def get_emote_sets(self, emote_set_ids: List[str]) -> Tuple[List[chat.Emote], str]:
        """
        Retrieve emotes for specific emote sets.

        Parameters
        ----------
        emote_set_ids: List[str]
            A list of emote set IDs to retrieve emotes for.

        Returns
        -------
        Tuple[List[chat.Emote], str]
            A tuple containing a list of dictionaries representing emotes and a template URL string.
        """
        data: Tuple[List[chat.Emote], str] = await self._connection.get_emote_sets(emote_set_ids)
        return data

    async def get_global_chat_badges(self) -> List[chat.Badge]:
        """
        Retrieve global chat badges.

        Returns
        -------
        List[chat.Badge]
            A list of dictionaries representing global chat badges.
        """
        data: List[chat.Badge] = await self._connection.get_global_chat_badges()
        return data

    async def fetch_channels_search(self,
                                    query: str,
                                    live_only: bool = False,
                                    first: int = 100) -> AsyncGenerator[List[search.ChannelSearch], None]:
        """
        Search for channels.

        Parameters
        ----------
        query: str
            The search query.
        live_only: bool
            Whether to return only live channels, by default False.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[search.ChannelSearch], None]
            A list of dictionaries representing channels matching the search query.
        """
        async for result in self._connection.fetch_channels_search(query, live_only, first):
            yield result

    async def fetch_streams(self,
                            user_logins: Optional[List[str]] = None,
                            user_ids: Optional[List[str]] = None,
                            category_ids: Optional[List[str]] = None,
                            stream_type: Literal['all', 'live'] = 'all',
                            language: Optional[str] = None,
                            first: int = 100) -> AsyncGenerator[List[streams.StreamInfo], None]:
        """
        Fetch streams based on various filters.

        Parameters
        ----------
        user_logins: Optional[List[str]]
            A list of user logins to filter streams by, by default None.
        user_ids: Optional[List[str]]
            A list of user IDs to filter streams by, by default None.
        category_ids: Optional[List[str]]
            A list of category IDs to filter streams by, by default None.
        stream_type: Literal['all', 'live']
            The type of streams to retrieve, by default 'all'.
        language: Optional[str]
            The language to filter streams by, by default None.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[streams.Stream], None]
            A list of dictionaries representing streams matching the filters.
        """
        async for result in self._connection.fetch_streams(user_logins,
                                                           user_ids,
                                                           category_ids,
                                                           stream_type,
                                                           language,
                                                           first):
            yield result

    async def fetch_videos_by_ids(self,
                                  video_ids: List[str],
                                  language: Optional[str] = None,
                                  period: Optional[Literal['all', 'day', 'month', 'week']] = None,
                                  sort: Optional[Literal['time', 'trending', 'views']] = None,
                                  video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']] = None,
                                  first: Optional[int] = 100) -> AsyncGenerator[List[channels.Video], None]:
        """
        Fetch videos by their IDs.

        Parameters
        ----------
        video_ids: List[str]
            A list of video IDs to retrieve.
        language: Optional[str]
            The language to filter videos by, by default None.
        period: Optional[Literal['all', 'day', 'month', 'week']]
            The period to filter videos by, by default None.
        sort: Optional[Literal['time', 'trending', 'views']]
            The sorting order for the videos, by default None.
        video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']]
            The type of videos to retrieve, by default None.
        first: Optional[int]
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[channels.Video], None]
            A list of dictionaries representing videos matching the filters.
        """
        async for result in self._connection.fetch_videos(None,
                                                          video_ids,
                                                          language,
                                                          period,
                                                          sort,
                                                          video_type,
                                                          first):
            yield result

    async def fetch_videos_by_category(self,
                                       category_id: str,
                                       language: Optional[str] = None,
                                       period: Optional[Literal['all', 'day', 'month', 'week']] = None,
                                       sort: Optional[Literal['time', 'trending', 'views']] = None,
                                       video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']] = None,
                                       first: Optional[int] = 100) -> AsyncGenerator[List[channels.Video], None]:
        """
        Fetch videos by category ID.

        Parameters
        ----------
        category_id: str
            The ID of the category to retrieve videos from.
        language: Optional[str]
            The language to filter videos by, by default None.
        period: Optional[Literal['all', 'day', 'month', 'week']]
            The period to filter videos by, by default None.
        sort: Optional[Literal['time', 'trending', 'views']]
            The sorting order for the videos, by default None.
        video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']]
            The type of videos to retrieve, by default None.
        first: Optional[int]
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[channels.Video], None]
            A list of dictionaries representing videos matching the filters.
        """
        async for result in self._connection.fetch_videos(category_id,
                                                          None,
                                                          language,
                                                          period,
                                                          sort,
                                                          video_type,
                                                          first):
            yield result

    async def fetch_clips_by_ids(self,
                                 clip_ids: List[str],
                                 started_at: Optional[datetime.datetime] = None,
                                 ended_at: Optional[datetime.datetime] = None,
                                 is_featured: Optional[bool] = None,
                                 first: int = 100) -> AsyncGenerator[List[channels.Clip], None]:
        """
        Fetch clips by their IDs.

        Parameters
        ----------
        clip_ids: List[str]
            A list of clip IDs to retrieve.
        started_at: Optional[datetime.datetime]
            The start time to filter clips by, by default None.
        ended_at: Optional[datetime.datetime]
            The end time to filter clips by, by default None.
        is_featured: Optional[bool]
            Whether to retrieve only featured clips, by default None.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[channels.Clip], None]
            A list of dictionaries representing clips matching the filters.
        """
        async for result in self._connection.fetch_clips(None,
                                                         clip_ids,
                                                         started_at,
                                                         ended_at,
                                                         is_featured,
                                                         first):
            yield result

    async def fetch_clips_by_category(self,
                                      category_id: str,
                                      started_at: Optional[datetime.datetime] = None,
                                      ended_at: Optional[datetime.datetime] = None,
                                      is_featured: Optional[bool] = None,
                                      first: int = 100) -> AsyncGenerator[List[channels.Clip], None]:
        """
        Fetch clips by category ID.

        Parameters
        ----------
        category_id: str
            The ID of the category to retrieve clips from.
        started_at: Optional[datetime.datetime]
            The start time to filter clips by, by default None.
        ended_at: Optional[datetime.datetime]
            The end time to filter clips by, by default None.
        is_featured: Optional[bool]
            Whether to retrieve only featured clips, by default None.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[channels.Clip], None]
            A list of dictionaries representing clips matching the filters.
        """
        async for result in self._connection.fetch_clips(category_id,
                                                         None,
                                                         started_at,
                                                         ended_at,
                                                         is_featured,
                                                         first):
            yield result

    async def get_content_classification_labels(self, locale: streams.Locale = 'en-US') -> List[streams.CCLInfo]:
        """
        Retrieve content classification labels.

        Parameters
        ----------
        locale: streams.Locale
            The locale to retrieve labels for, by default 'en-US'.

        Returns
        -------
        List[streams.CCLInfo]
            A list of dictionaries representing content classification labels.
        """
        data: List[streams.CCLInfo] = await self._connection.get_content_classification_labels(locale)
        return data

    async def get_global_cheermotes(self) -> List[bits.Cheermote]:
        """
        Retrieve global cheermotes.

        Returns
        -------
        List[bits.Cheermote]
            A list of dictionaries representing global cheermotes.
        """
        data: List[bits.Cheermote] = await self._connection.get_global_cheermotes()
        return data

    async def fetch_top_categories(self, first: int = 100) -> AsyncGenerator[List[search.Game], None]:
        """
        Fetch the top categories on Twitch.

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[search.Game], None]
            A list of dictionaries representing the top categories.
        """
        async for result in self._connection.fetch_top_games(first):
            yield result

    async def get_categories(self,
                             names: Optional[List[str]] = None,
                             ids: Optional[List[str]] = None,
                             igdb_ids: Optional[List[str]] = None) -> List[search.Game]:
        """
        Retrieve game categories by name, ID, or IGDB ID.

        Parameters
        ----------
        names: Optional[List[str]]
            A list of game names to filter by, by default None.
        ids: Optional[List[str]]
            A list of game IDs to filter by, by default None.
        igdb_ids: Optional[List[str]]
            A list of IGDB IDs to filter by, by default None.

        Returns
        -------
        List[search.Game]
            A list of dictionaries representing game categories.
        """
        data: List[search.Game] = await self._connection.get_games(game_names=names, game_ids=ids, igdb_ids=igdb_ids)
        return data

    async def fetch_categories_search(self, query: str,
                                      first: int = 100) -> AsyncGenerator[List[search.CategorySearch], None]:
        """
        Search for categories by a query string.

        Parameters
        ----------
        query: str
            The search query.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[search.CategorySearch], None]
            A list of dictionaries representing categories matching the search query.
        """
        async for result in self._connection.fetch_categories_search(query, first):
            yield result

    async def fetch_extension_analytics(self,
                                        extension_id: Optional[str] = None,
                                        analytics_type: Literal['overview_v2'] = 'overview_v2',
                                        started_at: Optional[datetime.datetime] = None,
                                        ended_at: Optional[datetime.datetime] = None,
                                        first: int = 100) -> AsyncGenerator[List[analytics.Extension], None]:
        """
        Fetch analytics data for extensions.

        Parameters
        ----------
        extension_id: Optional[str]
            The ID of the extension to filter by, by default None.
        analytics_type: Literal['overview_v2']
            The type of analytics data to retrieve, by default 'overview_v2'.
        started_at: Optional[datetime.datetime]
            The start time to filter analytics by, by default None.
        ended_at: Optional[datetime.datetime]
            The end time to filter analytics by, by default None.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[analytics.Extension], None]
            A list of dictionaries representing extension analytics data.
        """
        async for result in self._connection.fetch_extension_analytics(extension_id,
                                                                       analytics_type,
                                                                       started_at,
                                                                       ended_at,
                                                                       first):
            yield result

    async def fetch_game_analytics(self,
                                   game_id: Optional[str] = None,
                                   analytics_type: Literal['overview_v2'] = 'overview_v2',
                                   started_at: Optional[datetime.datetime] = None,
                                   ended_at: Optional[datetime.datetime] = None,
                                   first: int = 100) -> AsyncGenerator[List[analytics.Game], None]:
        """
        Fetch analytics data for games.

        Parameters
        ----------
        game_id: Optional[str]
            The ID of the game to filter by, by default None.
        analytics_type: Literal['overview_v2']
            The type of analytics data to retrieve, by default 'overview_v2'.
        started_at: Optional[datetime.datetime]
            The start time to filter analytics by, by default None.
        ended_at: Optional[datetime.datetime]
            The end time to filter analytics by, by default None.
        first: int
            The maximum number of results to retrieve, by default 100.

        Yields
        ------
        AsyncGenerator[List[analytics.Game], None]
            A list of dictionaries representing game analytics data.
        """
        async for result in self._connection.fetch_game_analytics(game_id,
                                                                  analytics_type,
                                                                  started_at,
                                                                  ended_at,
                                                                  first):
            yield result



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .channel import Channel, BroadcasterChannel, ClientChannel
from typing import TYPE_CHECKING, overload
from .utils import convert_rfc3339
from datetime import datetime

if TYPE_CHECKING:
    from .types import users, PData, chat, Data, PEdata, activity, channels, streams, TData
    from typing import Optional, Dict, Any, Tuple, List, AsyncGenerator, Literal, Union
    from .state import ConnectionState

__all__ = ('User', 'Broadcaster', 'ClientUser')


class BaseUser:
    """
    Represents base class of user.

    Attributes
    ----------
    id: str
        The unique identifier for the user.
    """
    __slots__ = ('id',)

    if TYPE_CHECKING:
        id: str

    def __init__(self, user_id: str) -> None:
        self._update(user_id=user_id)

    def __repr__(self) -> str:
        # Return a string representation of the BaseUser instance.
        return f'<BaseUser id={self.id}>'

    def __eq__(self, other: object) -> bool:
        """
        Check if this user is equal to another user by comparing IDs.

        Parameters
        ----------
        other: object
            The object to compare with.

        Returns
        -------
        bool
            True if the other object is a BaseUser with the same ID, False otherwise.
        """
        if isinstance(other, BaseUser):
            return self.id == other.id
        return False

    def __ne__(self, other: object) -> bool:
        """
        Check if this user is not equal to another user.

        Parameters
        ----------
        other: object
            The object to compare with.

        Returns
        -------
        bool
            True if the other object is not a BaseUser or has a different ID, False otherwise.
        """
        return not self.__eq__(other)

    def _update(self, user_id: str) -> None:
        # Update the user's ID.
        self.id = user_id


class User(BaseUser):
    """
    Represents a user with additional functionality.

    Attributes
    ----------
    id: str
        The unique identifier for the user.
    """
    __slots__ = ('_state',  '_auth_user_id', '__weakref__')

    def __init__(self, user_id: str, auth_user_id: str, *, state: ConnectionState) -> None:
        super().__init__(user_id=user_id)
        self._state: ConnectionState = state
        self._auth_user_id: str = auth_user_id

    @property
    def channel(self) -> Channel:
        """
        Get the channel associated with this user.

        Returns
        -------
        Channel
            The channel object representing the user's channel.
        """
        return Channel(self.id, self._auth_user_id, state=self._state)

    async def get_info(self) -> users.User:
        """
        Retrieve the full information about the user.

        Returns
        -------
        users.User
            A dictionary containing the user's information.
        """
        data: Data[List[users.User]] = await self._state.http.get_users(self._auth_user_id,
                                                                        user_ids=[self.id])
        return data['data'][0]

    async def get_chat_color(self) -> str:
        """
        Retrieve the chat color associated with the user.

        Returns
        -------
        str
            The hexadecimal color code representing the user's chat color.
        """
        data: Data[List[chat.UserChatColor]] = await self._state.http.get_user_chat_color(self._auth_user_id,
                                                                                          user_ids=[self.id])
        return data['data'][0]['color']


class Broadcaster(User):
    """
    Represents a broadcaster.

    Attributes
    ----------
    id: str
        The unique identifier for the broadcaster.
    """
    __slots__ = ()

    def __init__(self, user_id: str, *, state: ConnectionState) -> None:
        super().__init__(user_id, user_id, state=state)

    @property
    def channel(self) -> BroadcasterChannel:
        """
        Retrieve the broadcaster channel.

        Returns
        -------
        Optional[BroadcasterChannel]
            The channel of the current user.
        """
        return BroadcasterChannel(self.id, state=self._state)

    async def update(self, description: Optional[str]) -> users.User:
        """
        Update the user's profile information.

        | Scopes      | Description           |
        | ----------- | ----------------------|
        | `user:edit` | Manage a user object. |

        Parameters
        ----------
        description: Optional[str]
            The new description to set for the user, by default None.

        Returns
        -------
        users.User
            The updated `users.User` object with the new description.
        """
        data: Data[List[users.User]] = await self._state.http.update_user(self.id, description)
        return data['data'][0]

    async def fetch_emotes(self, user: Optional[User] = None) -> AsyncGenerator[Tuple[List[chat.Emote], str], None]:
        """
        Fetch emotes associated with the user or broadcaster.

        | Scopes            | Description                      |
        | ----------------- | ---------------------------------|
        | `user:read:emote` | View emotes available to a user. |

        Parameters
        ----------
        user: Optional[User]
            The user whose emotes to retrieve. If None, retrieves emotes for the broadcaster.

        Yields
        ------
        AsyncGenerator[Tuple[List[chat.Emote], str], None]
            A tuple where the first element is a list of `chat.Emote` dictionaries, and the second
            element is the template used for the emotes.
        """
        kwargs: Dict[str, Any] = {
            'user_id': self.id,
            'broadcaster_id': user.id if user else self.id,
            'after': None
        }
        while True:
            data: PEdata[List[chat.Emote]] = await self._state.http.get_user_emotes(**kwargs)
            yield data['data'], data['template']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_drops_entitlements(self,
                                       entitlement_ids: Optional[List[str]] = None,
                                       game_id: Optional[str] = None,
                                       fulfillment_status: Optional[Literal['claimed', 'fulfilled']] = None,
                                       first: int = 200) -> AsyncGenerator[List[activity.Entitlement], None]:
        """
        Fetch drop entitlements for the user.

        Parameters
        ----------
        entitlement_ids: Optional[List[str]]
            A list of entitlement IDs to filter by, by default None.
        game_id: Optional[str]
            The game ID to filter entitlements by, by default None.
        fulfillment_status: Optional[Literal['claimed', 'fulfilled']]
            The fulfillment status to filter by, by default None.
        first: int
            The maximum number of entitlements to retrieve per request, by default 200.

        Yields
        ------
        AsyncGenerator[List[activity.Entitlement], None]
            A list of `activity.Entitlement` dictionaries representing the entitlements.
        """
        kwargs: Dict[str, Any] = {
            'entitlement_ids': entitlement_ids,
            'user_id': self.id,
            'game_id': game_id,
            'fulfillment_status': fulfillment_status,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[activity.Entitlement]] = await self._state.http.get_drops_entitlements(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def update_drops_entitlements(self,
                                        entitlement_ids: List[str],
                                        fulfillment_status: Literal['claimed', 'fulfilled']
                                        ) -> List[activity.EntitlementsUpdate]:
        """
        Update the fulfillment status of drop entitlements.

        Parameters
        ----------
        entitlement_ids: List[str]
            A list of entitlement IDs to be updated.
        fulfillment_status: Literal['claimed', 'fulfilled']
            The new fulfillment status to set for the entitlements.

        Returns
        -------
        List[activity.EntitlementsUpdate]
            A list of `activity.EntitlementsUpdate` dictionaries representing the updated entitlements.
        """
        data: Data[List[activity.EntitlementsUpdate]] = await self._state.http.update_drops_entitlements(
            self.id,
            entitlement_ids,
            fulfillment_status)
        return data['data']

    async def update_chat_color(self, color: Union[str, chat.UserChatColors]) -> None:
        """
        Update the user's chat color.

        | Scopes                   | Description                                        |
        | ------------------------ | ---------------------------------------------------|
        | `user:manage:chat_color` | Update the color used for the user’s name in chat. |

        ???+ note
            To specify a color using a hexadecimal code, the user must have Turbo or Prime status.

        Parameters
        ----------
        color: Union[str, chat.UserChatColors]
            The new color to set for the user's chat messages. Can be a color code or a predefined color.
        """
        await self._state.http.update_user_chat_color(self.id, color)

    async def block(self,
                    user: User,
                    source_context: Literal['chat', 'whisper'] = None,
                    reason: Literal['harassment', 'spam', 'other'] = None) -> None:
        """
        Block a specified user.

        | Scopes                      | Description                      |
        | --------------------------- | ---------------------------------|
        | `user:manage:blocked_users` | Manage the block list of a user. |

        Parameters
        ----------
        user: User
            The user to block.
        source_context: Literal['chat', 'whisper']
            The context in which the block is applied, by default None.
        reason: Literal['harassment', 'spam', 'other']
            The reason for blocking the user, by default None.
        """
        await self._state.http.block_user(self.id, user.id, source_context, reason)

    async def unblock(self, user: User) -> None:
        """
        Unblock a previously blocked user.

        | Scopes                      | Description                      |
        | --------------------------- | ---------------------------------|
        | `user:manage:blocked_users` | Manage the block list of a user. |

        Parameters
        ----------
        user: User
            The user to unblock.
        """
        await self._state.http.unblock_user(self.id, user.id)

    async def fetch_blocked_users(self, first: int = 100) -> AsyncGenerator[List[users.SpecificUser], None]:
        """
        Fetch the list of users blocked by the broadcaster.

        | Scopes                    | Description                    |
        | ------------------------- | -------------------------------|
        | `user:read:blocked_users` | View the block list of a user. |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page, by default 100.

        Yields
        ------
        AsyncGenerator[List[users.SpecificUser], None]
            A list of dictionaries representing the blocked users.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self.id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[users.SpecificUser]] = await self._state.http.get_user_block_list(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def whisper(self, user: User, message: str) -> None:
        """
        Send a whisper message to a specified user.

        !!! Warning
            This action requires a verified phone number.

        | Scopes                 | Description                                                                  |
        | ---------------------- | -----------------------------------------------------------------------------|
        | `user:manage:whispers` | Receive whispers sent to your user, and send whispers on your user’s behalf. |

        Parameters
        ----------
        user: User
            The user to whom the whisper message will be sent.
        message: str
            The content of the whisper message.
        """
        await self._state.http.send_whisper(self.id, user.id, message)

    @overload
    async def check_followed(self, user: User) -> channels.Follows:
        ...

    @overload
    async def check_followed(self, user: User) -> None:
        ...

    async def check_followed(self, user: User) -> Optional[channels.Follows]:
        """
        Check if the broadcaster is following a specified user.

        | Scopes              | Description                               |
        | ------------------- | ------------------------------------------|
        | `user:read:follows` | View the list of channels a user follows. |

        Parameters
        ----------
        user: User
            The user to check for a follow relationship.

        Returns
        -------
        Optional[channels.Follows]
            A dictionary if the broadcaster follows the specified user; otherwise, None.
        """
        data: TData[List[channels.Follows]] = await self._state.http.get_followed_channels(self.id,
                                                                                           broadcaster_id=user.id)
        return data['data'][0] if len(data['data']) != 0 else None

    async def get_total_followed(self) -> int:
        """
        Retrieve the total number of channels followed by the broadcaster.

        | Scopes              | Description                               |
        | ------------------- | ------------------------------------------|
        | `user:read:follows` | View the list of channels a user follows. |

        Returns
        -------
        int
            The total number of channels followed by the broadcaster.
        """
        data: TData[List[channels.Follows]] = await self._state.http.get_followed_channels(self.id, first=1)
        return data['total']

    async def fetch_followed(self, first: int = 100) -> AsyncGenerator[List[channels.Follows], None]:
        """
        Fetch the list of channels followed by the broadcaster.

        | Scopes              | Description                               |
        | ------------------- | ------------------------------------------|
        | `user:read:follows` | View the list of channels a user follows. |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page, by default 100.

        Yields
        ------
        AsyncGenerator[List[channels.Follows], None]
            A list of dictionaries representing the followed channels.
        """
        kwargs: Dict[str, Any] = {
            'user_id': self.id,
            'first': first,
            'after': None
        }
        while True:
            data: TData[List[channels.Follows]] = await self._state.http.get_followed_channels(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_followed_streaming(self, first: int = 100) -> AsyncGenerator[List[streams.StreamInfo], None]:
        """
        Fetch the list of currently streaming channels followed by the broadcaster.

        | Scopes              | Description                               |
        | ------------------- | ------------------------------------------|
        | `user:read:follows` | View the list of channels a user follows. |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page, by default 100.

        Yields
        ------
        AsyncGenerator[List[streams.Stream], None]
            A list of dictionaries representing the streams from followed channels.
        """
        kwargs: Dict[str, Any] = {
            'user_id': self.id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[streams.StreamInfo]] = await self._state.http.get_followed_streams(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    @overload
    async def check_user_subscription(self, user: User) -> channels.SubscriptionCheck:
        ...

    @overload
    async def check_user_subscription(self, user: User) -> None:
        ...

    async def check_user_subscription(self, user: User) -> Optional[channels.SubscriptionCheck]:
        """
        Check if the broadcaster is subscribed to the specified user.

        | Scopes                    | Description                                                    |
        | ------------------------- | ---------------------------------------------------------------|
        | `user:read:subscriptions` | View if an authorized user is subscribed to specific channels. |

        Parameters
        ----------
        user: User
            The user to check for a subscription relationship.

        Returns
        -------
        Optional[channels.SubscriptionCheck]
            A dictionary representing if the broadcaster is subscribed to the specified user; otherwise, None.
        """
        data: Data[List[channels.SubscriptionCheck]] = await self._state.http.check_user_subscription(self.id, user.id)
        return data['data'][0] if len(data['data']) != 0 else None

    async def fetch_moderated_channels(self, first: int = 100) -> AsyncGenerator[List[users.Broadcaster], None]:
        """
        Fetch the list of channels moderated by the broadcaster.

        | Scopes                         | Description                                                 |
        | ------------------------------ | ------------------------------------------------------------|
        | `user:read:moderated_channels` | Read the list of channels you have moderator privileges in. |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page, by default 100.

        Yields
        ------
        AsyncGenerator[List[users.Broadcaster], None]
            A list of dictionaries representing the channels moderated by the broadcaster.
        """
        kwargs: Dict[str, Any] = {
            'user_id': self.id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[users.Broadcaster]] = await self._state.http.get_moderated_channels(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break


class ClientUser(Broadcaster):
    """
    Represents a user associated with the client.

    !!! Danger
        The attributes are read-only.
        These attributes are automatically updated via EventSub whenever user information changes.

    Attributes
    ----------
    id: str
        The unique identifier for the user.
    name: str
        The login name of the user.
    display_name: str
        The display name of the user.
    description: str
        A brief description or bio of the user.
    email: Optional[str]
        The email address of the user, if available and accessible. Requires `user:read:email` scope.
    joined_at: datetime
        The date and time when the user joined Twitch.
    """

    __slots__ = ('_channel_data', 'name', 'display_name', 'description', 'joined_at', 'email')

    def __init__(self,
                 *,
                 state: ConnectionState,
                 user_data: users.User,
                 channel_data: channels.ChannelInfo,
                 ) -> None:
        super().__init__(user_data['id'], state=state)
        self._channel_data: channels.ChannelInfo = channel_data
        self.name: str = user_data['login']
        self.display_name: str = user_data['display_name']
        self.description: str = user_data['description']
        self.joined_at: datetime = convert_rfc3339(user_data['created_at'])
        # Requires user:read:email scope
        self.email: Optional[str] = user_data.get('email') or None

    @property
    def channel(self) -> ClientChannel:
        """
        Retrieve the client-user channel.

        Returns
        -------
        Optional[broadcasterChannel]
            The channel of the current user.
        """
        return ClientChannel(self.id, state=self._state, data=self._channel_data)

    async def update(self, description: Optional[str]) -> users.User:
        """
        Update the user's profile information.

        | Scopes      | Description           |
        | ----------- | ----------------------|
        | `user:edit` | Manage a user object. |

        Parameters
        ----------
        description: Optional[str]
            The new description to set for the user, by default None.

        Returns
        -------
        users.User
            The updated `users.User` object with the new description.
        """
        data: Data[List[users.User]] = await self._state.http.update_user(self.id, description)
        self.description = data['data'][0]['description']
        return data['data'][0]



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .utils import datetime_to_str, convert_rfc3339
from .stream import Stream, BroadcasterStream
from typing import overload, TYPE_CHECKING
from .chat import Chat, BroadcasterChat

if TYPE_CHECKING:
    from .types import (Data, PData, TData, channels, moderation, streams, users, TPData, activity, DTData, bits,
                        interaction)
    from typing import Optional, List, AsyncGenerator, Literal, Dict, Any
    from .state import ConnectionState
    from .user import User
    import datetime

__all__ = ('Channel', 'BroadcasterChannel', 'ClientChannel')


class Channel:
    """
    Represents a channel.
    """

    __slots__ = ('_state', '_user_id', '_auth_user_id')

    def __init__(self, user_id: str, auth_user_id: str, *, state: ConnectionState) -> None:
        self._state: ConnectionState = state
        self._user_id: str = user_id
        self._auth_user_id: str = auth_user_id

    @property
    def chat(self) -> Chat:
        """
        Get the chat functionality associated with this channel.

        Returns
        -------
        Chat
            The `Chat` object representing the chat functionality for the channel.
        """
        return Chat(self._user_id, self._auth_user_id, state=self._state)

    @property
    def stream(self) -> Stream:
        """
        Get the streaming functionality associated with this channel.

        Returns
        -------
        Stream
            The `Stream` object representing the streaming functionality for the channel.
        """
        return Stream(self._user_id, self._auth_user_id, state=self._state)

    async def get_info(self) -> channels.ChannelInfo:
        """
        Retrieve detailed information about the channel.

        Returns
        -------
        channels.ChannelInfo
            A dictionary containing the channel's information, including details such as title, description, and more.
        """
        data: Data[List[channels.ChannelInfo]] = await self._state.http.get_channel_information(
            self._auth_user_id, [self._user_id])
        return data['data'][0]

    async def get_teams(self) -> List[channels.ChannelTeam]:
        """
        Retrieve the teams associated with the channel.

        Returns
        -------
        List[dict]
            A list of dictionaries representing the teams that the channel is part of.
        """
        data: Data[List[channels.ChannelTeam]] = await self._state.http.get_channel_teams(self._auth_user_id,
                                                                                          broadcaster_id=self._user_id)
        return data['data']

    async def get_total_followers(self) -> int:
        """
        Retrieve the total number of followers for the channel.

        | Scopes                     | Description                          |
        | -------------------------- | -------------------------------------|
        | `moderator:read:followers` | Read the followers of a broadcaster. |

        Returns
        -------
        int
            The total count of followers for the channel.
        """
        data: TData[List[channels.Follower]] = await self._state.http.get_channel_followers(
            self._auth_user_id,
            broadcaster_id=self._user_id,
            first=1)
        return data['total']

    @overload
    async def check_follower(self, user: User) -> channels.Follower:
        ...

    @overload
    async def check_follower(self, user: User) -> None:
        ...

    async def check_follower(self, user: User) -> Optional[channels.Follower]:
        """
        Check if a specific user is a follower of the channel.

        | Scopes                     | Description                          |
        | -------------------------- | -------------------------------------|
        | `moderator:read:followers` | Read the followers of a broadcaster. |

        Parameters
        ----------
        user: User
            The user to check for being a follower.

        Returns
        -------
        Optional[channels.Follower]
            A dictionary representing the follower if the user is a follower, otherwise `None`.
        """
        data: TData[List[channels.Follower]] = await self._state.http.get_channel_followers(self._auth_user_id,
                                                                                            self._user_id,
                                                                                            user_id=user.id)
        return data['data'][0] if len(data['data']) != 0 else None

    async def fetch_followers(self, *, first: int = 100) -> AsyncGenerator[List[channels.Follower], None]:
        """
        Retrieve followers for the channel in pages.

        | Scopes                     | Description                          |
        | -------------------------- | -------------------------------------|
        | `moderator:read:followers` | Read the followers of a broadcaster. |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[channels.Follower], None]
            A generator yielding lists of dictionaries representing followers.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: TData[List[channels.Follower]] = await self._state.http.get_channel_followers(self._auth_user_id,
                                                                                                **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def get_banned_users(self, __users: List[User], /) -> List[moderation.BannedUser]:
        """
        Retrieve information about banned users from the channel.

        | Scopes                          | Description                       |
        | ------------------------------- | ----------------------------------|
        | `moderation:read`               | View a channel’s moderation data. |
        | `moderator:manage:banned_users` | Ban and unban users.              |

        Parameters
        ----------
        __users: List[User]
            A list of users to check if they are banned.

        Returns
        -------
        List[moderation.BannedUser]
            A list of dictionaries representing the banned users.
        """
        data: PData[List[moderation.BannedUser]] = await self._state.http.get_banned_users(
            self._auth_user_id,
            self._user_id,
            user_ids=[user.id for user in __users])
        return data['data']

    async def fetch_banned_users(self, *, first: int = 100) -> AsyncGenerator[List[moderation.BannedUser], None]:
        """
        Retrieve banned users from the channel in pages.

        | Scopes                          | Description                                                       |
        | ------------------------------- | ------------------------------------------------------------------|
        | `moderation:read`               | View a channel’s moderation data including Moderators, Bans, ..., |
        | `moderator:manage:banned_users` | Ban and unban users.                                              |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[moderation.BannedUser], None]
            A generator yielding lists of dictionaries representing banned users.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[moderation.BannedUser]] = await self._state.http.get_banned_users(self._auth_user_id,
                                                                                               **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def ban(self, user: User, duration: Optional[int] = None, reason: Optional[str] = None) -> None:
        """
        Ban a user from the channel.

        | Scopes                          | Description          |
        | ------------------------------- | ---------------------|
        | `moderator:manage:banned_users` | Ban and unban users. |

        Parameters
        ----------
        user: User
            The user to be banned.
        duration: Optional[int]
            The duration of the ban in minutes.
        reason: Optional[str]
            The reason for banning the user.
        """
        await self._state.http.ban_user(self._user_id, self._auth_user_id, user.id, duration=duration, reason=reason)

    async def unban(self, user: User) -> None:
        """
        Unban a user from the channel.

        | Scopes                          | Description          |
        | ------------------------------- | ---------------------|
        | `moderator:manage:banned_users` | Ban and unban users. |

        Parameters
        ----------
        user: User
            The user to be unbanned.
        """
        await self._state.http.unban_user(self._user_id, self._auth_user_id, user.id)

    @overload
    async def check_unban_request(self,
                                  user: User,
                                  *,
                                  status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
                                  ) -> moderation.UnBanRequest:
        ...

    @overload
    async def check_unban_request(self,
                                  user: User,
                                  *,
                                  status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
                                  ) -> None:
        ...

    async def check_unban_request(self,
                                  user: User,
                                  *,
                                  status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
                                  ) -> Optional[moderation.UnBanRequest]:
        """
        Retrieve a specific unban request for a user.

        | Scopes                            | Description                            |
        | --------------------------------- | ---------------------------------------|
        | `moderator:read:unban_requests`   | View a broadcaster’s unban requests.   |
        | `moderator:manage:unban_requests` | Manage a broadcaster’s unban requests. |

        Parameters
        ----------
        user: User
            The user for whom the unban request is to be retrieved.
        status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
            The status of the unban request to filter by.

        Returns
        -------
        Optional[moderation.UnBanRequest]
            A dictionary containing the details of the unban request if it exists; otherwise, None.
        """
        data: PData[List[moderation.UnBanRequest]] = await self._state.http.get_unban_requests(self._user_id,
                                                                                               self._auth_user_id,
                                                                                               status=status,
                                                                                               user_id=user.id)
        return data['data'][0] if len(data['data']) != 0 else None

    async def fetch_unban_requests(self,
                                   status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled'],
                                   *,
                                   first: int = 100) -> AsyncGenerator[List[moderation.UnBanRequest], None]:
        """
        Retrieve unban requests from the channel in pages.

        | Scopes                            | Description                            |
        | --------------------------------- | ---------------------------------------|
        | `moderator:read:unban_requests`   | View a broadcaster’s unban requests.   |
        | `moderator:manage:unban_requests` | Manage a broadcaster’s unban requests. |

        Parameters
        ----------
        status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
            The status of the unban requests to filter by.
        first: int
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[moderation.UnBanRequest], None]
            A generator yielding lists of dictionaries representing unban requests.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'moderator_id': self._auth_user_id,
            'status': status.lower(),
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[moderation.UnBanRequest]] = await self._state.http.get_unban_requests(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def resolve_unban_request(self,
                                    request_id: str,
                                    status: Literal['approved', 'denied'],
                                    resolution_text: Optional[str] = None) -> moderation.UnBanRequest:
        """
        Resolve an unban request.

        | Scopes                            | Description                            |
        | --------------------------------- | ---------------------------------------|
        | `moderator:manage:unban_requests` | Manage a broadcaster’s unban requests. |

        Parameters
        ----------
        request_id: str
            The ID of the unban request to resolve.
        status: Literal['approved', 'denied']
            The status to set for the unban request.
        resolution_text: Optional[str]
            Optional text explaining the resolution.

        Returns
        -------
        moderation.UnBanRequest
            A dictionary containing the details of the resolved unban request.
        """
        data: Data[List[moderation.UnBanRequest]] = await self._state.http.resolve_unban_requests(
            self._user_id,
            self._auth_user_id,
            unban_request_id=request_id,
            status=status,
            resolution_text=resolution_text)
        return data['data'][0]

    async def fetch_recent_stream_markers(self, first: int = 100) -> AsyncGenerator[List[streams.StreamMarker], None]:
        """
        Retrieve recent stream markers for the channel in pages.

        | Scopes                     | Description                                 |
        | -------------------------- | --------------------------------------------|
        | `user:read:broadcast`      | View a user’s broadcasting configuration.   |
        | `channel:manage:broadcast` | Manage a channel’s broadcast configuration. |

        Parameters
        ----------
        first: int
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[streams.StreamMarker], None]
            A generator yielding lists of dictionaries representing stream markers.
        """
        kwargs: Dict[str, Any] = {
            'user_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[streams.StreamMarker]] = await self._state.http.get_stream_markers(self._auth_user_id,
                                                                                                **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_video_markers(self,
                                  video_id: str,
                                  first: int = 100) -> AsyncGenerator[List[streams.StreamMarker], None]:
        """
        Retrieve markers for a specific video in pages.

        | Scopes                     | Description                                 |
        | -------------------------- | --------------------------------------------|
        | `user:read:broadcast`      | View a user’s broadcasting configuration.   |
        | `channel:manage:broadcast` | Manage a channel’s broadcast configuration. |

        Parameters
        ----------
        video_id: str
            The ID of the video for which to retrieve markers.
        first: int
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[streams.StreamMarker], None]
            A generator yielding lists of dictionaries representing video markers.
        """
        kwargs: Dict[str, Any] = {
            'video_id': video_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[streams.StreamMarker]] = await self._state.http.get_stream_markers(self._auth_user_id,
                                                                                                **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_videos(self,
                           language: Optional[str] = None,
                           period: Optional[Literal['all', 'day', 'month', 'week']] = None,
                           sort: Optional[Literal['time', 'trending', 'views']] = None,
                           video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']] = None,
                           first: Optional[int] = 100) -> AsyncGenerator[List[channels.Video], None]:
        """
        Retrieve videos for the channel in pages.

        Parameters
        ----------
        language: Optional[str]
            The language of the videos to filter by, or None to include all languages.
        period: Optional[Literal['all', 'day', 'month', 'week']]
            The time period for the videos, or None to include all periods.
        sort: Optional[Literal['time', 'trending', 'views']]
            The sorting method for the videos, or None to use the default sort.
        video_type: Optional[Literal['all', 'archive', 'highlight', 'upload']]
            The type of videos to retrieve, or None to include all types.
        first: Optional[int]
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[channels.Video], None]
            A generator yielding lists of dictionaries representing videos.
        """
        kwargs: Dict[str, Any] = {
            'user_id': self._user_id,
            'language': language,
            'period': period,
            'sort': sort,
            'video_type': video_type,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[channels.Video]] = await self._state.http.get_videos(self._auth_user_id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_clips(self,
                          started_at: Optional[datetime] = None,
                          ended_at: Optional[datetime] = None,
                          is_featured: Optional[bool] = None,
                          first: int = 100
                          ) -> AsyncGenerator[List[channels.Clip], None]:
        """
        Retrieve clips from the channel in pages.

        Parameters
        ----------
        started_at: Optional[datetime]
            The start date and time for filtering clips.
        ended_at: Optional[datetime]
            The end date and time for filtering clips.
        is_featured: Optional[bool]
            Whether to filter by featured clips.
        first: int
            The maximum number of results to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[channels.Clip], None]
            A generator yielding lists of dictionaries representing clips.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'started_at': datetime_to_str(started_at),
            'ended_at': datetime_to_str(ended_at),
            'is_featured': is_featured,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[channels.Clip]] = await self._state.http.get_clips(self._auth_user_id, **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break


class BroadcasterChannel(Channel):
    """
    Represents a broadcaster channel.
    """

    __slots__ = ()

    def __init__(self, user_id: str, *, state: ConnectionState) -> None:
        super().__init__(user_id, user_id, state=state)

    @property
    def chat(self) -> BroadcasterChat:
        """
        Get the chat functionality associated with broadcaster channel.

        Returns
        -------
        BroadcasterChat
            The `BroadcasterChat` object representing the chat functionality for the channel.
        """
        return BroadcasterChat(self._user_id, state=self._state)

    @property
    def stream(self) -> BroadcasterStream:
        """
        Get the streaming functionality associated with broadcaster channel.

        Returns
        -------
        BroadcasterStream
            The `BroadcasterStream` object representing the streaming functionality for the channel.
        """
        return BroadcasterStream(self._user_id, state=self._state)

    async def get_extensions(self) -> List[channels.Extension]:
        """
        Retrieve all extensions associated with the user.

        | Scopes                | Description                                        |
        | --------------------- | ---------------------------------------------------|
        | `user:read:broadcast` | View a user’s broadcasting configuration.          |
        | `user:edit:broadcast` | View and edit a user’s broadcasting configuration. |

        Returns
        -------
        List[channels.Extension]
            A list of dictionaries representing the user's extensions.
        """
        data: Data[List[channels.Extension]] = await self._state.http.get_user_extensions(self._user_id)
        return data['data']

    async def get_active_extensions(self) -> channels.ActiveExtensions:
        """
        Retrieve all active extensions associated with the user.

        | Scopes                     | Description                                         |
        | --------------------------- | ---------------------------------------------------|
        | `user:read:broadcast`       | View a user’s broadcasting configuration.          |
        | `channel:manage:extensions` | Manage a channel’s Extension configuration.        |
        | `user:edit:broadcast`       | View and edit a user’s broadcasting configuration. |

        Returns
        -------
        channels.ActiveExtensions
            A dictionary representing the currently active extensions.
        """
        data: Data[channels.ActiveExtensions] = await self._state.http.get_user_active_extensions(self._user_id)
        return data['data']

    @overload
    async def update_extension(self,
                               key: Literal['panel'],
                               *,
                               number: Literal['1', '2', '3'],
                               extension_id: str,
                               extension_version: str,
                               activate: bool,
                               x: None = None,
                               y: None = None) -> channels.ActiveExtensions:
        ...

    @overload
    async def update_extension(self,
                               key: Literal['overlay'],
                               *,
                               number: Literal[1],
                               extension_id: str,
                               extension_version: str,
                               activate: bool,
                               x: None = None,
                               y: None = None) -> channels.ActiveExtensions:
        ...

    @overload
    async def update_extension(self,
                               key: Literal['component'],
                               *,
                               number: Literal['1', '2'],
                               extension_id: str,
                               extension_version: str,
                               activate: bool,
                               x: int,
                               y: int) -> channels.ActiveExtensions:
        ...

    async def update_extension(self,
                               key: Literal['panel', 'overlay', 'component'],
                               /,
                               **kwargs: Any) -> channels.ActiveExtensions:
        """
        Update an extension based on the provided parameters.

        | Scopes                     | Description                                         |
        | --------------------------- | ---------------------------------------------------|
        | `channel:manage:extensions` | Manage a channel’s Extension configuration.        |
        | `user:edit:broadcast`       | View and edit a user’s broadcasting configuration. |

        Parameters
        ----------
        key: Literal['panel', 'overlay', 'component']
            The type of extension to update, which can be 'panel', 'overlay', or 'component'.
        **kwargs: dict
            Keyword arguments for the extension update. The available arguments depend on the `key` value:
                - `number` (Literal): The number of the extension to update.
                - `extension_id` (str): The ID of the extension to update.
                - `extension_version` (str): The version of the extension to update.
                - `activate` (bool): Whether to activate or deactivate the extension.
                - `x` (int): The x-coordinate position for 'component' extensions.
                - `y` (int): The y-coordinate position for 'component' extensions.

        Returns
        -------
        channels.ActiveExtensions
            A dictionary representing the updated active extensions.
        """
        data: Data[channels.ActiveExtensions] = await self._state.http.update_user_extensions(self._user_id,
                                                                                              key=key,
                                                                                              **kwargs)
        return data['data']

    async def update(self,
                     category_id: Optional[str] = None,
                     broadcaster_language: Optional[str] = None,
                     title: Optional[str] = None,
                     delay: Optional[int] = None,
                     tags: Optional[List[str]] = None,
                     content_classification_labels: Optional[List[channels.CCL]] = None,
                     is_branded_content: Optional[bool] = None) -> None:
        """
        Update the channel information.

        | Scopes                     | Description                                 |
        | -------------------------- | --------------------------------------------|
        | `channel:manage:broadcast` | Manage a channel’s broadcast configuration. |

        Parameters
        ----------
        category_id: Optional[str]
            The ID of the game or category.
        broadcaster_language: Optional[str]
            The language of the broadcaster.
        title: Optional[str]
            The title of the stream.
        delay: Optional[int]
            The delay in seconds for the stream.
        tags: Optional[List[str]]
            A list of tags for the stream.
        content_classification_labels: Optional[List[channels.CCL]]
            A list of content classification labels.
        is_branded_content: Optional[bool]
            Whether the content is branded.
        """
        await self._state.http.modify_channel_information(self._user_id, category_id, broadcaster_language,
                                                          title, delay, tags, content_classification_labels,
                                                          is_branded_content)

    async def get_moderators(self, __users: List[User], /) -> List[users.SpecificUser]:
        """
        Retrieve a list of specific moderators for the channel.

        | Scopes                       | Description                                                 |
        | --------------------------- | -------------------------------------------------------------|
        | `moderation:read`           | View a channel’s moderation data.                            |
        | `channel:manage:moderators` | Add or remove the moderator role from users in your channel. |

        Parameters
        ----------
        __users: List[User]
            A list of `User` objects representing the users whose moderation status is being queried.

        Returns
        -------
        List[users.SpecificUser]
            A list of dictionaries representing the specific moderators for the channel.
        """
        data: PData[List[users.SpecificUser]] = await self._state.http.get_moderators(self._user_id,
                                                                                      user_ids=[u.id for u in __users],
                                                                                      first=100)
        return data['data']

    async def fetch_moderators(self, first: int = 100) -> AsyncGenerator[List[users.SpecificUser], None]:
        """
        Fetch all moderators for the channel.

        | Scopes                       | Description                                                 |
        | --------------------------- | -------------------------------------------------------------|
        | `moderation:read`           | View a channel’s moderation data.                            |
        | `channel:manage:moderators` | Add or remove the moderator role from users in your channel. |

        Parameters
        ----------
        first: int
            The number of moderators to fetch per request.

        Yields
        ------
        AsyncGenerator[List[users.SpecificUser], None]
            An async generator yielding lists of dictionaries representing the moderators for the channel.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[users.SpecificUser]] = await self._state.http.get_moderators(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def add_moderator(self, user: User) -> None:
        """
        Add a user as a moderator for the channel.

        | Scopes                       | Description                                                 |
        | --------------------------- | -------------------------------------------------------------|
        | `channel:manage:moderators` | Add or remove the moderator role from users in your channel. |

        Parameters
        ----------
        user: User
            The `User` object representing the user to be added as a moderator.
        """
        await self._state.http.add_channel_moderator(self._user_id, user.id)

    async def remove_moderator(self, user: User) -> None:
        """
        Remove a user from the list of moderators for the channel.

        | Scopes                       | Description                                                 |
        | --------------------------- | -------------------------------------------------------------|
        | `channel:manage:moderators` | Add or remove the moderator role from users in your channel. |


        Parameters
        ----------
        user: User
            The `User` object representing the user to be removed as a moderator.
        """
        await self._state.http.remove_channel_moderator(self._user_id, user.id)

    async def get_editors(self) -> List[channels.Editor]:
        """
        Retrieve all editors associated with the user's channel.

        | Scopes                 | Description                                              |
        | ---------------------- | ---------------------------------------------------------|
        | `channel:read:editors` | View a list of users with the editor role for a channel. |

        Returns
        -------
        List[channels.Editor]
            A list of dictionaries representing the editors of the channel.
        """
        data: Data[List[channels.Editor]] = await self._state.http.get_channel_editors(self._user_id)
        return data['data']

    async def get_vips(self, __users: List[User], /) -> List[users.SpecificUser]:
        """
        Retrieve specific VIPs from the user's channel.

        | Scopes                | Description                                            |
        | --------------------- | -------------------------------------------------------|
        | `channel:read:vips`   | Read the list of VIPs in your channel.                 |
        | `channel:manage:vips` | Add or remove the VIP role from users in your channel. |

        Parameters
        ----------
        __users: List[User]
            A list of users to check for VIP status.

        Returns
        -------
        List[users.SpecificUser]
            A list of dictionaries representing the specific VIPs in the channel.
        """
        data: PData[List[users.SpecificUser]] = await self._state.http.get_vips(self._user_id,
                                                                                user_ids=[u.id for u in __users],
                                                                                first=100)
        return data['data']

    async def fetch_vips(self, first: int = 100) -> AsyncGenerator[List[users.SpecificUser], None]:
        """
        Fetch all VIPs for the channel.

        | Scopes                | Description                                            |
        | --------------------- | -------------------------------------------------------|
        | `channel:read:vips`   | Read the list of VIPs in your channel.                 |
        | `channel:manage:vips` | Add or remove the VIP role from users in your channel. |

        Parameters
        ----------
        first: int
            The number of VIPs to fetch per request.

        Yields
        ------
        AsyncGenerator[List[users.SpecificUser], None]
            An async generator yielding lists of dictionaries representing the VIPs for the channel.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[users.SpecificUser]] = await self._state.http.get_vips(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def add_vip(self, user: User) -> None:
        """
        Add a user as a VIP in the channel.

        | Scopes                | Description                                            |
        | --------------------- | -------------------------------------------------------|
        | `channel:manage:vips` | Add or remove the VIP role from users in your channel. |

        Parameters
        ----------
        user: User
            The user to be added as a VIP.
        """
        await self._state.http.add_channel_vip(self._user_id, user.id)

    async def remove_vip(self, user: User) -> None:
        """
        Remove a user's VIP status in the channel.

        | Scopes                | Description                                            |
        | --------------------- | -------------------------------------------------------|
        | `channel:manage:vips` | Add or remove the VIP role from users in your channel. |

        Parameters
        ----------
        user: User
            The user whose VIP status should be removed.
        """
        await self._state.http.remove_channel_vip(self._user_id, user.id)

    async def get_total_subscriptions(self) -> int:
        """
        Retrieve the total number of subscriptions for the broadcaster.

        | Scopes                      | Description                                   |
        | ---------------------------- | ---------------------------------------------|
        | `channel:read:subscriptions` | View a list of all subscribers to a channel. |

        Returns
        -------
        int
            The total number of subscriptions for the broadcaster.
        """
        data: TPData[List[channels.Subscription]] = await self._state.http.get_broadcaster_subscriptions(self._user_id,
                                                                                                         first=1)
        return data['total']

    async def get_subscription_points(self) -> int:
        """
        Retrieve the total subscription points for the broadcaster.

        | Scopes                      | Description                                   |
        | ---------------------------- | ---------------------------------------------|
        | `channel:read:subscriptions` | View a list of all subscribers to a channel. |

        Returns
        -------
        int
            The total subscription points for the broadcaster.
        """
        data: TPData[List[channels.Subscription]] = await self._state.http.get_broadcaster_subscriptions(self._user_id,
                                                                                                         first=1)
        return data['points']

    async def get_subscriptions(self, __users: List[User], /) -> List[channels.Subscription]:
        """
        Retrieve the subscription details for specific users.

        | Scopes                      | Description                                   |
        | ---------------------------- | ---------------------------------------------|
        | `channel:read:subscriptions` | View a list of all subscribers to a channel. |

        Parameters
        ----------
        __users: List[User]
            A list of users whose subscription details are to be retrieved.

        Returns
        -------
        List[channels.Subscription]
            A list of dictionaries representing the subscription details of the specified users.
        """
        data: TPData[List[channels.Subscription]] = await self._state.http.get_broadcaster_subscriptions(
            self._user_id,
            user_ids=[u.id for u in __users],
            first=100)
        return data['data']

    async def fetch_subscriptions(self, first: int = 100) -> AsyncGenerator[List[channels.Subscription], None]:
        """
        Fetch all subscriptions for the broadcaster.

        | Scopes                       | Description                                  |
        | ---------------------------- | ---------------------------------------------|
        | `channel:read:subscriptions` | View a list of all subscribers to a channel. |

        Parameters
        ----------
        first: int
            The maximum number of subscriptions to retrieve per request.

        Yields
        ------
        AsyncGenerator[List[channels.Subscription], None]
            A list of dictionaries representing the subscriptions for each page.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: TPData[List[channels.Subscription]] = await self._state.http.get_broadcaster_subscriptions(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    @overload
    async def get_goals(self) -> activity.Goal:
        ...

    @overload
    async def get_goals(self) -> None:
        ...

    async def get_goals(self) -> Optional[activity.Goal]:
        """
        Retrieve the current creator goals for the broadcaster.

        | Scopes               | Description                       |
        | -------------------- | ----------------------------------|
        | `channel:read:goals` | View Creator Goals for a channel. |

        Returns
        -------
        Optional[activity.Goal]
            A dictionary representing the current creator goal, or None if no goals are active.
        """
        data: Data[List[activity.Goal]] = await self._state.http.get_creator_goals(self._user_id)
        return data['data'][0] if len(data['data']) != 0 else None

    async def get_bits_leaderboard(self,
                                   period: Optional[Literal['day', 'week', 'month', 'year', 'all']] = None,
                                   started_at: Optional[datetime] = None,
                                   user: Optional[User] = None,
                                   count: int = 100) -> List[bits.Leaderboard]:
        """
        Retrieve the bits leaderboard for the broadcaster.

        | Scopes      | Description                          |
        | ----------- | -------------------------------------|
        | `bits:read` | View Bits information for a channel. |

        Parameters
        ----------
        period: Optional[Literal['day', 'week', 'month', 'year', 'all']]
            The period for which to retrieve the leaderboard.
        started_at: Optional[datetime]
            The start date for the leaderboard.
        user: Optional[User]
            A specific user to filter the leaderboard.
        count: int
            The number of leaderboard entries to retrieve.

        Returns
        -------
        List[bits.Leaderboard]
            A list of dictionaries representing the bits leaderboard entries.
        """
        kwargs: Dict[str, Any] = {
            'count': count,
            'period': period,
            'started_at': convert_rfc3339(started_at),
            'user_id': user.id if user is not None else None
        }
        data: DTData[List[bits.Leaderboard]] = await self._state.http.get_bits_leaderboard(self._user_id, **kwargs)
        return data['data']

    @overload
    async def get_charity_campaign(self) -> activity.Charity:
        ...

    @overload
    async def get_charity_campaign(self) -> None:
        ...

    async def get_charity_campaign(self) -> Optional[activity.Charity]:
        """
        Retrieve the current charity campaign for the broadcaster.

        | Scopes                 | Description                                                       |
        | ---------------------- | ------------------------------------------------------------------|
        | `channel:read:charity` | Read charity campaign details and user donations on your channel. |

        Returns
        -------
        Optional[activity.Charity]
            A dictionary representing the current charity campaign, or None if no campaign is active.
        """
        data: Data[List[activity.Charity]] = await self._state.http.get_charity_campaign(self._user_id)
        return data['data'][0] if len(data['data']) != 0 else None

    async def fetch_charity_donations(self, first: int = 100) -> AsyncGenerator[List[activity.CharityDonation], None]:
        """
        Fetch all donations for the current charity campaign.

        | Scopes                 | Description                                                       |
        | ---------------------- | ------------------------------------------------------------------|
        | `channel:read:charity` | Read charity campaign details and user donations on your channel. |

        Parameters
        ----------
        first: int
            The maximum number of donations to retrieve per request.

        Yields
        ------
        AsyncGenerator[List[activity.CharityDonation], None]
            A list of dictionaries representing the charity donations for each page.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[activity.CharityDonation]] = await self._state.http.get_charity_campaign_donations(
                **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def fetch_hype_trains(self, first: int = 100) -> AsyncGenerator[List[interaction.HypeTrain], None]:
        """
        Fetch all Hype Train events for the broadcaster.

        | Scopes                    | Description                                |
        | ------------------------- | -------------------------------------------|
        | `channel:read:hype_train` | View Hype Train information for a channel. |

        Parameters
        ----------
        first: int
            The maximum number of Hype Train events to retrieve per request.

        Yields
        ------
        AsyncGenerator[List[interaction.HypeTrain], None]
            A list of dictionaries representing the Hype Train events for each page.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[interaction.HypeTrain]] = await self._state.http.get_hype_train_events(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def get_rewards(self,
                          reward_ids: Optional[List[str]] = None,
                          only_manageable_rewards: bool = False) -> List[interaction.Reward]:
        """
        Retrieve a list of custom rewards for the broadcaster.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:read:redemptions`   | View Channel Points custom rewards and their redemptions on a channel.   |
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        reward_ids: Optional[List[str]]
            A list of reward IDs to filter the results.
        only_manageable_rewards: bool
            Whether to return only rewards that can be managed by the broadcaster.

        Returns
        -------
        List[interaction.Reward]
            A list of dictionaries representing the custom rewards.
        """
        data: Data[List[interaction.Reward]] = await self._state.http.get_custom_reward(
            self._user_id,
            reward_ids,
            only_manageable_rewards)
        return data['data']

    async def create_reward(self,
                            title: str,
                            cost: int,
                            is_enabled: bool = True,
                            background_color: Optional[str] = None,
                            is_user_input_required: bool = False,
                            prompt: Optional[str] = None,
                            is_max_per_stream_enabled: bool = False,
                            max_per_stream: Optional[int] = None,
                            is_max_per_user_per_stream_enabled: bool = False,
                            max_per_user_per_stream: Optional[int] = None,
                            is_global_cooldown_enabled: bool = False,
                            global_cooldown_seconds: Optional[int] = None,
                            should_redemptions_skip_request_queue: bool = False) -> interaction.Reward:
        """
        Create a new custom reward for the broadcaster.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        title: str
            The title of the reward.
        cost: int
            The cost of the reward in channel points.
        is_enabled: bool
            Whether the reward is enabled.
        background_color: Optional[str]
            The background color of the reward.
        is_user_input_required: bool
            Whether user input is required to redeem the reward.
        prompt: Optional[str]
            The prompt to display for the reward.
        is_max_per_stream_enabled: bool
            Whether there is a limit on the number of times the reward can be redeemed per stream.
        max_per_stream: Optional[int]
            The maximum number of times the reward can be redeemed per stream.
        is_max_per_user_per_stream_enabled: bool
            Whether there is a limit on the number of times a user can redeem the reward per stream.
        max_per_user_per_stream: Optional[int]
            The maximum number of times a user can redeem the reward per stream.
        is_global_cooldown_enabled: bool
            Whether there is a global cooldown between redemptions.
        global_cooldown_seconds: Optional[int]
            The cooldown time in seconds between redemptions.
        should_redemptions_skip_request_queue: bool
            Whether redemptions should skip the request queue.

        Returns
        -------
        interaction.Reward
            A dictionary representing the created reward.
        """
        data: Data[List[interaction.Reward]] = await self._state.http.create_custom_rewards(
            self._user_id,
            title,
            cost,
            is_enabled,
            background_color,
            is_user_input_required,
            prompt,
            is_max_per_stream_enabled,
            max_per_stream,
            is_max_per_user_per_stream_enabled,
            max_per_user_per_stream,
            is_global_cooldown_enabled,
            global_cooldown_seconds,
            should_redemptions_skip_request_queue)
        return data['data'][0]

    async def update_reward(self,
                            reward_id: str,
                            title: Optional[str] = None,
                            cost: Optional[int] = None,
                            is_enabled: Optional[bool] = None,
                            background_color: Optional[str] = None,
                            is_user_input_required: Optional[bool] = None,
                            prompt: Optional[str] = None,
                            is_max_per_stream_enabled: Optional[bool] = None,
                            max_per_stream: Optional[int] = None,
                            is_max_per_user_per_stream_enabled: Optional[bool] = None,
                            max_per_user_per_stream: Optional[int] = None,
                            is_global_cooldown_enabled: Optional[bool] = None,
                            global_cooldown_seconds: Optional[int] = None,
                            should_redemptions_skip_request_queue: Optional[bool] = None) -> interaction.Reward:
        """
        Update an existing custom reward for the broadcaster.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        reward_id: str
            The ID of the reward to update.
        title: Optional[str]
            The new title of the reward.
        cost: Optional[int]
            The new cost of the reward in channel points.
        is_enabled: Optional[bool]
            Whether the reward is enabled.
        background_color: Optional[str]
            The new background color of the reward.
        is_user_input_required: Optional[bool]
            Whether user input is required to redeem the reward.
        prompt: Optional[str]
            The new prompt to display for the reward.
        is_max_per_stream_enabled: Optional[bool]
            Whether there is a limit on the number of times the reward can be redeemed per stream.
        max_per_stream: Optional[int]
            The new maximum number of times the reward can be redeemed per stream.
        is_max_per_user_per_stream_enabled: Optional[bool]
            Whether there is a limit on the number of times a user can redeem the reward per stream.
        max_per_user_per_stream: Optional[int]
            The new maximum number of times a user can redeem the reward per stream.
        is_global_cooldown_enabled: Optional[bool]
            Whether there is a global cooldown between redemptions.
        global_cooldown_seconds: Optional[int]
            The new cooldown time in seconds between redemptions.
        should_redemptions_skip_request_queue: Optional[bool]
            Whether redemptions should skip the request queue.

        Returns
        -------
        interaction.Reward
            A dictionary representing the updated reward.
        """
        data: Data[List[interaction.Reward]] = await self._state.http.update_custom_reward(
            self._user_id,
            reward_id,
            title,
            cost,
            is_enabled,
            background_color,
            is_user_input_required,
            prompt,
            is_max_per_stream_enabled,
            max_per_stream,
            is_max_per_user_per_stream_enabled,
            max_per_user_per_stream,
            is_global_cooldown_enabled,
            global_cooldown_seconds,
            should_redemptions_skip_request_queue)
        return data['data'][0]

    async def delete_reward(self, reward_id: str) -> None:
        """
        Delete an existing custom reward for the broadcaster.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        reward_id: str
            The ID of the reward to delete.
        """
        await self._state.http.delete_custom_reward(self._user_id, reward_id)

    async def get_reward_redemptions(self,
                                     reward_id: str,
                                     redemptions_ids: List[str],
                                     sort: Literal['oldest', 'newest'] = 'oldest'
                                     ) -> List[interaction.RewardRedemption]:
        """
        Retrieve specific redemptions for a custom reward.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:read:redemptions`   | View Channel Points custom rewards and their redemptions on a channel.   |
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        reward_id: str
            The ID of the reward for which to retrieve redemptions.
        redemptions_ids: List[str]
            A list of redemption IDs to filter the results.
        sort: Literal['oldest', 'newest']
            The order in which to sort the redemptions.

        Returns
        -------
        List[interaction.RewardRedemption]
            A list of dictionaries representing the redemptions.
        """
        data: PData[List[interaction.RewardRedemption]] = await self._state.http.get_custom_reward_redemption(
            self._user_id,
            reward_id,
            redemptions_ids,
            sort=sort,
            first=50)
        return data['data']

    async def fetch_reward_redemptions(self,
                                       reward_id: str,
                                       status: Literal['canceled', 'fulfilled', 'unfulfilled'],
                                       sort: Literal['oldest', 'newest'] = 'oldest',
                                       first: int = 50
                                       ) -> AsyncGenerator[List[interaction.RewardRedemption], None]:
        """
        Fetch redemptions for a custom reward.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:read:redemptions`   | View Channel Points custom rewards and their redemptions on a channel.   |
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        reward_id: str
            The ID of the reward for which to fetch redemptions.
        status: Literal['canceled', 'fulfilled', 'unfulfilled']
            The status of the redemptions to fetch.
        sort: Literal['oldest', 'newest']
            The order in which to sort the redemptions.
        first: int
            The maximum number of redemptions to retrieve per request.

        Yields
        ------
        AsyncGenerator[List[interaction.RewardRedemption], None]
            A list of dictionaries representing the redemptions for each page.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'reward_id': reward_id,
            'status': status,
            'sort': sort,
            'first': first,
            'after': None,
        }
        while True:
            data: PData[List[interaction.RewardRedemption]] = await self._state.http.get_custom_reward_redemption(
                **kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def update_reward_redemptions(self,
                                        reward_id: str,
                                        redemptions_ids: List[str],
                                        *,
                                        status: Literal['canceled', 'fulfilled']
                                        ) -> List[interaction.RewardRedemption]:
        """
        Update the status of specific reward redemptions.

        | Scopes                       | Description                                                              |
        | ---------------------------- | -------------------------------------------------------------------------|
        | `channel:manage:redemptions` | Manage Channel Points custom rewards and their redemptions on a channel. |

        Parameters
        ----------
        reward_id: str
            The ID of the reward for which to update redemptions.
        redemptions_ids: List[str]
            A list of redemption IDs to update.
        status: Literal['canceled', 'fulfilled']
            The new status of the redemptions.

        Returns
        -------
        List[interaction.RewardRedemption]
            A list of dictionaries representing the updated redemptions.
        """
        data: Data[List[interaction.RewardRedemption]] = await self._state.http.update_redemption_status(
            self._user_id,
            reward_id,
            redemptions_ids,
            status=status)
        return data['data']

    async def get_predictions(self, prediction_ids: List[str]) -> List[interaction.Prediction]:
        """
        Retrieve a list of predictions by their IDs.

        | Scopes                       | Description                                     |
        | ---------------------------- | ------------------------------------------------|
        | `channel:read:predictions`   | View a channel’s Channel Points Predictions.    |
        | `channel:manage:predictions` | Manage of channel’s Channel Points Predictions. |

        Parameters
        ----------
        prediction_ids: List[str]
            A list of prediction IDs to retrieve.

        Returns
        -------
        List[interaction.Prediction]
            A list of dictionaries representing the predictions.
        """
        data: PData[List[interaction.Prediction]] = await self._state.http.get_predictions(
            self._user_id,
            prediction_ids,
            first=25
        )
        return data['data']

    async def fetch_predictions(self, first: int = 25) -> AsyncGenerator[List[interaction.Prediction], None]:
        """
        Fetch broadcaster predictions.

        | Scopes                       | Description                                     |
        | ---------------------------- | ------------------------------------------------|
        | `channel:read:predictions`   | View a channel’s Channel Points Predictions.    |
        | `channel:manage:predictions` | Manage of channel’s Channel Points Predictions. |

        Parameters
        ----------
        first: int
            The maximum number of predictions to retrieve per request.

        Yields
        ------
        AsyncGenerator[List[interaction.Prediction], None]
            A list of dictionaries representing the predictions for each page.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[interaction.Prediction]] = await self._state.http.get_predictions(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def create_prediction(self,
                                title: str,
                                outcomes: List[str],
                                window: int) -> interaction.Prediction:
        """
        Create a new prediction for the broadcaster.

        | Scopes                       | Description                                     |
        | ---------------------------- | ------------------------------------------------|
        | `channel:manage:predictions` | Manage of channel’s Channel Points Predictions. |

        Parameters
        ----------
        title: str
            The title of the prediction.
        outcomes: List[str]
            A list of possible outcomes for the prediction.
        window: int
            The duration of the prediction window in seconds.

        Returns
        -------
        interaction.Prediction
            A dictionary representing the created prediction.
        """
        data: Data[List[interaction.Prediction]] = await self._state.http.create_prediction(self._user_id,
                                                                                            title,
                                                                                            outcomes,
                                                                                            window)
        return data['data'][0]

    @overload
    async def end_prediction(self,
                             prediction_id: str,
                             status: Literal['resolved'],
                             winning_outcome_id: str
                             ) -> interaction.Prediction:
        ...

    @overload
    async def end_prediction(self,
                             prediction_id: str,
                             status: Literal['canceled', 'locked'],
                             winning_outcome_id: None = None
                             ) -> interaction.Prediction:
        ...

    async def end_prediction(self,
                             prediction_id: str,
                             status: Any[Literal['resolved', 'canceled', 'locked']],
                             winning_outcome_id: Optional[str] = None) -> interaction.Prediction:
        """
        End an existing prediction with a specified status.

        | Scopes                       | Description                                     |
        | ---------------------------- | ------------------------------------------------|
        | `channel:manage:predictions` | Manage of channel’s Channel Points Predictions. |

        Parameters
        ----------
        prediction_id: str
            The ID of the prediction to end.
        status: Literal['resolved', 'canceled', 'locked']
            The status to set for the prediction.
        winning_outcome_id: Optional[str]
            The ID of the winning outcome if the status is 'resolved'.

        Returns
        -------
        interaction.Prediction
            A dictionary representing the updated prediction.
        """
        data: Data[List[interaction.Prediction]] = await self._state.http.end_prediction(self._user_id,
                                                                                         prediction_id,
                                                                                         status,
                                                                                         winning_outcome_id)
        return data['data'][0]

    async def get_polls(self, poll_ids: List[str]) -> List[interaction.Poll]:
        """
        Retrieve a list of polls by their IDs.

        | Scopes                 | Description               |
        | ---------------------- | --------------------------|
        | `channel:read:polls`   | View a channel’s polls.   |
        | `channel:manage:polls` | Manage a channel’s polls. |

        Parameters
        ----------
        poll_ids: List[str]
            A list of poll IDs to retrieve.

        Returns
        -------
        List[interaction.Poll]
            A list of dictionaries representing the polls.
        """
        data: PData[List[interaction.Poll]] = await self._state.http.get_polls(self._user_id,
                                                                               poll_ids,
                                                                               first=20)
        return data['data']

    async def fetch_polls(self, first: int = 20) -> AsyncGenerator[List[interaction.Poll], None]:
        """
        Fetch broadcaster polls.

        | Scopes                 | Description               |
        | ---------------------- | --------------------------|
        | `channel:read:polls`   | View a channel’s polls.   |
        | `channel:manage:polls` | Manage a channel’s polls. |

        Parameters
        ----------
        first: int
            The maximum number of polls to retrieve per request.

        Yields
        ------
        AsyncGenerator[List[interaction.Poll], None]
            A list of dictionaries representing the polls for each page.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[interaction.Poll]] = await self._state.http.get_polls(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def create_poll(self,
                          title: str,
                          choices: List[str],
                          duration: int,
                          channel_points_voting_enabled: bool = False,
                          channel_points_per_vote: Optional[int] = None) -> interaction.Poll:
        """
        Create a new poll for the broadcaster.

        | Scopes                 | Description               |
        | ---------------------- | --------------------------|
        | `channel:manage:polls` | Manage a channel’s polls. |

        Parameters
        ----------
        title: str
            The title of the poll.
        choices: List[str]
            A list of choices for the poll.
        duration: int
            The duration of the poll in seconds.
        channel_points_voting_enabled: bool
            Whether channel points voting is enabled.
        channel_points_per_vote: Optional[int]
            The number of channel points required per vote if channel points voting is enabled.

        Returns
        -------
        interaction.Poll
            A dictionary representing the created poll.
        """
        data: Data[List[interaction.Poll]] = await self._state.http.create_poll(self._user_id,
                                                                                title,
                                                                                choices,
                                                                                duration,
                                                                                channel_points_voting_enabled,
                                                                                channel_points_per_vote)
        return data['data'][0]

    async def end_poll(self, poll_id: str, status: Literal['terminated', 'archived']) -> interaction.Poll:
        """
        End an existing poll with a specified status.

        | Scopes                 | Description               |
        | ---------------------- | --------------------------|
        | `channel:manage:polls` | Manage a channel’s polls. |

        Parameters
        ----------
        poll_id: str
            The ID of the poll to end.
        status: Literal['terminated', 'archived']
            The status to set for the poll, either 'terminated' or 'archived'.

        Returns
        -------
        interaction.Poll
            A dictionary representing the updated poll.
        """
        data: Data[List[interaction.Poll]] = await self._state.http.end_poll(self._user_id, poll_id, status)
        return data['data'][0]

    async def delete_videos(self, video_ids: List[str]) -> List[str]:
        """
        Delete a list of videos by their IDs.

        | Scopes                  | Description                |
        | ----------------------- | ---------------------------|
        | `channel:manage:videos` | Manage a channel’s videos. |

        Parameters
        ----------
        video_ids: List[str]
            A list of video IDs to delete.

        Returns
        -------
        List[str]
            A list of IDs of the deleted videos.
        """
        data: Data[List[str]] = await self._state.http.delete_videos(self._user_id, video_ids)
        return data['data']


class ClientChannel(BroadcasterChannel):
    """
    Represents a client-user channel.

    !!! Danger
        The attributes are read-only.
        These attributes are automatically updated via EventSub whenever channel information changes.

    Attributes
    ----------
    title: str
        The title of the channel.
    language: str
        The language of the broadcaster's content.
    category_id: str
        The ID of the category (game) associated with the channel.
    category_name: str
        The name of the category (game) associated with the channel.
    ccl: List[channels.CCL]
        A list of content classification labels associated with the channel.
    """

    __slots__ = ('title', 'language', 'category_id', 'category_name', 'ccl')

    def __init__(self, user_id: str, /, state: ConnectionState, data: channels.ChannelInfo) -> None:
        super().__init__(user_id, state=state)
        self._state: ConnectionState = state
        self.title: str = data['title']
        self.language: str = data['broadcaster_language']
        self.category_id: str = data['game_id']
        self.category_name: str = data['game_name']
        self.ccl: List[channels.CCL] = data['content_classification_labels']



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


class CheerEvent(SpecificBroadcaster):
    """
    Represents a channel cheer event on Twitch.

    Attributes
    ----------
    is_anonymous: bool
        Indicates whether the cheer was made anonymously.
    user_id: Optional[str]
        The ID of the user who made the cheer, if not anonymous.
    user_login: Optional[str]
        The login name of the user who made the cheer, if not anonymous.
    user_name: Optional[str]
        The display name of the user who made the cheer, if not anonymous.
    message: str
        The message accompanying the cheer.
    bits: int
        The number of bits cheered.
    """
    is_anonymous: bool
    user_id: Optional[str]
    user_login: Optional[str]
    user_name: Optional[str]
    message: str
    bits: int



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificUser, Broadcaster, Moderator
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import Optional, Literal


# Automod
class AutoModMessageStatus(TypedDict):
    """
    Represents the status of an AutoMod message.

    Attributes
    ----------
    msg_id: str
        The unique identifier for the message.
    is_permitted: bool
        Whether the message was permitted by AutoMod.
    """
    msg_id: str
    is_permitted: bool


class AutoModSettings(TypedDict):
    """
    Represents AutoMod settings.

    Attributes
    ----------
    broadcaster_id: str
        The ID of the broadcaster.
    moderator_id: str
        The ID of the moderator.
    overall_level: Optional[int]
        The overall level of AutoMod moderation.
    disability: int
        The level of moderation for disability-related content.
    aggression: int
        The level of moderation for aggressive content.
    sexuality_sex_or_gender: int
        The level of moderation for content related to sexuality, sex, or gender.
    misogyny: int
        The level of moderation for misogynistic content.
    bullying: int
        The level of moderation for bullying-related content.
    swearing: int
        The level of moderation for swearing.
    race_ethnicity_or_religion: int
        The level of moderation for content related to race, ethnicity, or religion.
    sex_based_terms: int
        The level of moderation for sex-based terms.
    """
    broadcaster_id: str
    moderator_id: str
    overall_level: Optional[int]
    disability: int
    aggression: int
    sexuality_sex_or_gender: int
    misogyny: int
    bullying: int
    swearing: int
    race_ethnicity_or_religion: int
    sex_based_terms: int


# Ban / Unban
class BannedUser(SpecificUser, Moderator):
    """
    Represents a banned user.

    Attributes
    ----------
    reason: str
        The reason the user was banned.
    created_at: str
        The timestamp when the ban was created.
    expires_at: str
        The timestamp when the ban expires.
    """
    reason: str
    created_at: str
    expires_at: str


class BanUser(TypedDict):
    """
    Represents a request to ban a user.

    Attributes
    ----------
    broadcaster_id: str
        The ID of the broadcaster.
    moderator_id: str
        The ID of the moderator issuing the ban.
    user_id: str
        The ID of the user to be banned.
    created_at: str
        The timestamp when the ban was created.
    end_time: str
        The timestamp when the ban will end.
    """
    broadcaster_id: str
    moderator_id: str
    user_id: str
    created_at: str
    end_time: str


class UnBanRequest(Broadcaster, Moderator, SpecificUser):
    """
    Represents a request to unban a user.

    Attributes
    ----------
    id: str
        The unique identifier for the unban request.
    text: str
        The text of the unban request.
    status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
        The status of the unban request.
    created_at: str
        The timestamp when the request was created.
    resolved_at: Optional[str]
        The timestamp when the request was resolved.
    resolution_text: Optional[str]
        The resolution text of the request, if applicable.
    """
    id: str
    text: str
    status: Literal['pending', 'approved', 'denied', 'acknowledged', 'canceled']
    created_at: str
    resolved_at: Optional[str]
    resolution_text: Optional[str]


# BlockedTerms
class BlockedTerm(TypedDict):
    """
    Represents a blocked term.

    Attributes
    ----------
    broadcaster_id: str
        The ID of the broadcaster.
    moderator_id: str
        The ID of the moderator.
    id: str
        The unique identifier for the blocked term.
    text: str
        The text of the blocked term.
    created_at: str
        The timestamp when the term was blocked.
    updated_at: str
        The timestamp when the term was last updated.
    expires_at: Optional[str]
        The timestamp when the block will expire, if applicable.
    """
    broadcaster_id: str
    moderator_id: str
    id: str
    text: str
    created_at: str
    updated_at: str
    expires_at: Optional[str]


# ShieldMode
class ShieldModeStatus(TypedDict):
    """
    Represents the status of Shield Mode.

    Attributes
    ----------
    is_active: bool
        Indicates if Shield Mode is active.
    moderator_id: str
        The ID of the moderator who activated Shield Mode.
    moderator_login: str
        The login name of the moderator who activated Shield Mode.
    moderator_name: str
        The display name of the moderator who activated Shield Mode.
    last_activated_at: str
        The timestamp when Shield Mode was last activated.
    """
    is_active: bool
    moderator_id: str
    moderator_login: str
    moderator_name: str
    last_activated_at: str


# Warning
class UserWarningResponse(TypedDict):
    """
    Represents a warning issued to a user.

    Attributes
    ----------
    broadcaster_id: str
        The ID of the broadcaster issuing the warning.
    user_id: str
        The ID of the user receiving the warning.
    moderator_id: str
        The ID of the moderator issuing the warning.
    reason: str
        The reason for the warning.
    """
    broadcaster_id: str
    user_id: str
    moderator_id: str
    reason: str
